/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed May  1 01:24:49 2002 by Jeff Dalton
 * Copyright: (c) 2000, AIAI, University of Edinburgh
 */

package ix.iface.domain;

import java.io.*;
import java.util.*;

import ix.icore.*;
import ix.icore.domain.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

/**
 * A parser for domains described in a Lisp-like syntax. <p>
 *
 * The parser parses a file of refinement definitions to populate 
 * a Domain. <p>
 *
 * In the syntax definitions below, upper-case names are literals
 * and lower-case names are syntactic variables.  The literals
 * can be written in either case in domain definitions.
 *
 * <pre>
 *    (REFINEMENT name pattern clause*)
 *
 *    name ::= symbol | string
 *    pattern ::= (item*)
 *    item ::= number | symbol | string | pattern | variable
 *    variable ::= ?symbol
 *    clause ::= (ISSUES issue*)
 *            | (NODES node*)
 *            | (ORDERINGS ordering*)
 *            | (CONSTRAINTS constraint*)
 *    issue ::= (ISSUE pattern)
 *    node ::= (node-id pattern)
 *    ordering ::= ({node-id | (node-id*)}*)
 *    constraint ::= (WORLD-STATE CONDITION pattern = value)
 *                |  (WORLD-STATE EFFECT pattern = value)
 *    value ::= item
 * </pre>
 *
 * @see Domain
 * @see #expandAllOrderings(LList)
 */

public class LTF_Parser extends DomainParser implements LTF_Symbols {

    protected String filename;
    protected LispReader lin;

    protected Object lastRead = "Start of file";

    public LTF_Parser(String filename) throws FileNotFoundException {
	this.filename = filename;
	this.lin = new LispFileReader(filename);
	addConstraintParsers();
    }

    public LTF_Parser(File file) throws FileNotFoundException {
	this(file.getPath());
    }

    public LTF_Parser() {
	// Used e.g. if someone just wants to call the constraint parser.
	addConstraintParsers();
    }

    public Domain readDomain() {
	return readDomain(new Domain());
    }

    public Domain readDomain(Domain dom) {
	Debug.noteln("Reading definitions from", filename);
	try {
	    Object item;
	    while ((item = lin.readObject()) != Lisp.EOF) {
		if (!(item instanceof LList)) {
		    syntaxError("Found " + item
				+ " when expecting a refinement definition");
		    continue;	// doesn't actually get here /\/
		}
		LList def = (LList)item;
		lastRead = item;
		Symbol defForm = (Symbol)def.elementAt(0);
		if (defForm == S_REFINEMENT) {
		    Refinement r = null;
		    try { r = makeRefinement(def); }
		    catch (SyntaxException e) {
			Debug.noteln("Rethrowing syntax exception.");
			e.fillInStackTrace();
			throw e;
		    }
		    catch (Exception e) {
			Debug.noteln("Problem with " + def);
			Debug.noteException(e);
			syntaxError(e.toString());
			continue; // doesn't actually get here /\/
		    }
		    dom.addRefinement(r);
		}
		else {
		    syntaxError("Illegal definition type " + defForm);
		}
	    }
	}
	catch (LispReadException e) {
	    Debug.noteException(e, false);
	    syntaxError(e.toString());
	}
	return dom;
    }

    protected void syntaxError(String message) {
	throw new SyntaxException(message
				  // + " while parsing " + Util.quote(filename)
				  + " in or after " + lastRead);
    }

    /*
     * Refinements
     */

    public Refinement makeRefinement(LList def) {
	Debug.assert(def.get(0) == S_REFINEMENT);

	Refinement r = new Refinement();
	r.setName(def.get(1).toString());
	r.setPattern((LList)def.get(2));
	for (LList body = def.drop(3); !body.isEmpty(); body = body.cdr()) {
	    Cons clause = (Cons)body.car();
	    Symbol name = (Symbol)clause.car();
	    LList items = clause.cdr();
	    if (name == S_NODES)
		r.setNodes(makeNodes(items));
	    else if (name == S_ORDERINGS)
		r.setOrderings(makeOrderings(items));
	    else if (name == S_CONSTRAINTS)
		r.setConstraints(makeConstraints(items));
	    else if (name == S_ISSUES)
		r.setIssues(makeIssues(items));
	    else
		syntaxError("Invalid clause " + clause);
	}
	return r;
    }

    public LList makeNodes(LList items) {
	LListCollector result = new LListCollector();
	for (Iterator i = items.iterator(); i.hasNext();) {
	    LList pair = (Cons)i.next();
	    Name id = Name.valueOf(pair.get(0));
	    LList pattern = (LList)pair.get(1);
	    result.add(new NodeSpec(id, pattern));
	}
	return result.isEmpty() ? null : result.contents();
    }

    public LList makeOrderings(LList items) {
	LListCollector result = new LListCollector();
	for (LList ords = expandAllOrderings(items);
	     !ords.isEmpty(); ords = ords.cdr()) {
	    Cons pair = (Cons)ords.car();
	    Name from = Name.valueOf(pair.get(0));
	    Name to = Name.valueOf(pair.get(1));
	    result.add(new Ordering(new NodeEndRef(End.END, from),
				    new NodeEndRef(End.BEGIN, to)));
	}
	return result.isEmpty() ? null : result.contents();
    }

    public LList makeConstraints(LList items) {
	LListCollector result = new LListCollector();
	for (Iterator i = items.iterator(); i.hasNext();) {
	    Object spec = i.next();
	    Constraint c = (Constraint)parsers.match(spec);
	    if (c == null)
		syntaxError("Invalid constraint " + spec);
	    else
		result.add(c);
	}
	return result.isEmpty() ? null : result.contents();
    }

    public LList makeIssues(LList items) {
	LListCollector result = new LListCollector();
	for (Iterator i = items.iterator(); i.hasNext();) {
	    LList issue = (LList)i.next();
	    if (issue.get(0) != S_ISSUE)
		syntaxError("Invalid issue syntax in " + issue);
	    else {
		LList pattern = (LList)issue.get(1);
		result.add(new Issue(pattern));
	    }
	}
	return result.isEmpty() ? null : result.contents();
    }


    /*
     * Orderings
     */

    /**
     * Expands a list of orderings into a list of simple orderings.
     * Each simple ordering is a pair of node ids: (before after). <p>
     *
     * Each ordering is a list representing a sequence.  Consider
     * adjacent elements A and B in such a list.  Each node in A is
     * linked before each node in B.  For example, (1 (2 3) 4) is
     * equivalent to the four separate lists (1 2), (1 3), (2 4), (3 4).
     */
    public static LList expandAllOrderings(LList orderings) {
	LListCollector result = new LListCollector();
	for (LList ords = orderings; ords != Lisp.NIL; ords = ords.cdr()) {
	    result.concLList(expandOrdering((LList)ords.car()));
	}
	return result.contents();
    }

    public static LList expandOrdering(LList ordering) {
	LListCollector result = new LListCollector();
	// for adjacent elements left and right ...
	for (LList ords = ordering; ords != Lisp.NIL; ords = ords.cdr()) {
	    Object left = ords.car();
	    Object right = ords.cdr().car();
	    result.concLList(expandOrderPair(left, right));
	}
	return result.contents();
    }

    static LList expandOrderPair(Object left, Object right) {
	LListCollector result = new LListCollector();
	for (LList l = ensureList(left); l != Lisp.NIL; l = l.cdr()) {
	    for (LList r = ensureList(right); r != Lisp.NIL; r = r.cdr()) {
		result.addElement(Lisp.list(l.car(), r.car()));
	    }
	}
	return result.contents();
    }

    static LList ensureList(Object a) {
	return a instanceof LList ? (LList)a : Lisp.list(a);
    }


    /*
     * Constraint parsing
     */

    protected MatchTable parsers = new MatchTable();

    public Constraint parseConstraint(Object spec) {
	// May be call from outside the LTF parser.
	// /\/ Which is why it doesn't call syntaxError.
	return (Constraint)parsers.match(spec);
    }


    protected static abstract class ConstraintParser extends MatchCase {
	ConstraintParser(String pat) {
	    this.pattern = Lisp.readFromString(pat);
	}
	public Object tryMatch(Object spec) {
	    Debug.noteln("Matching against", pattern);
	    return SimpleMatcher.match(pattern, spec);
	}
	public Object ifSelected(Object spec, Object match) {
	    return makeConstraint((LList)spec, (MatchEnv)match);
	}
	public abstract Constraint makeConstraint(LList spec, MatchEnv env);
    }

    protected void addConstraintParsers() {

        parsers.addCase(
	    new ConstraintParser(
	        "(world-state condition ?pattern = ?value)") {

	    public Constraint makeConstraint(LList spec, MatchEnv env) {
		LList pat = (LList)env.get(Q_PATTERN);
		Object val = env.get(Q_VALUE);
		return new Constraint(
		    (Symbol)spec.get(0), 	// world-state
		    (Symbol)spec.get(1),	// condition
		    Lisp.list(
		        new PatternAssignment(pat, val)));
	    }
	});

        parsers.addCase(
	    new ConstraintParser(
	        "(world-state effect ?pattern = ?value)") {

	    public Constraint makeConstraint(LList spec, MatchEnv env) {
		LList pat = (LList)env.get(Q_PATTERN);
		Object val = env.get(Q_VALUE);
		return new Constraint(
		    (Symbol)spec.get(0), 	// world-state
		    (Symbol)spec.get(1),	// effect
		    Lisp.list(
		        new PatternAssignment(pat, val)));
	    }
	});

    }

}

// Issues:
// * Add schemas one at a time or pass a list to new Domain()?
// * Tell the domain to analyse itself or leave it "raw"?
// * If we gave the domain a list of all schemas, it could
//   do the analysis automaticlly.  When we give it schemas
//   one at a time, it needs to be told it has them all.
// * Should the constructor take a String or a File?
// * Need to do more syntax-checking.
