/****************************************************************************
 * A simple button panel with buttons of the same size.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui;

import java.util.*;
import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.BoxLayout;
import java.awt.event.MouseListener;
import java.awt.GridLayout;

/**
 * A simple button panel with buttons of the same size.
 *
 * The button panel has an orientation (BoxLayout.X_AXIS or BoxLayout.Y_AXIS)
 * and any number of buttons whose strings are given at creation or with
 * a call to addButton. All buttons in the panel will have the same size.
 */
public class IXButPanel extends JPanel {
  public boolean groupFlag;
  public ButtonGroup group;
  public HashSet buttons = new HashSet();

    /** Creates a panel for buttons with Y_AXIS layout. */
    public IXButPanel(){
        this(BoxLayout.Y_AXIS);
        }
    /**
     * Creates a panel for buttons with the given axis.
     * @param axis the layout orientation: 
     *    BoxLayout.X_AXIS or BoxLayout.Y_AXIS
     */
    public IXButPanel(int axis){
        super(); 
        if ((axis == BoxLayout.Y_AXIS)||(axis == BoxLayout.X_AXIS)){
            this.setLayout(new BoxLayout(this, axis));}
    }
    /**
     * Creates a panel with given buttons, given axis, adding the given
     * listener.
     *
     * @param mouseListener a listener interested in clicks of the buttons
     * @param axis the layout orientation: 
     *    BoxLayout.X_AXIS or BoxLayout.Y_AXIS
     * @param buttons a string array of button names
     */
    public IXButPanel(MouseListener mouseListener, int axis, Object[] buttons){
        super();    
        int x; int y;
        if (axis == BoxLayout.X_AXIS) {x=buttons.length; y=1;}
        else {x=0; y=buttons.length;}
        this.setLayout(new GridLayout(y,x));
        for (int i=0; i<buttons.length; i++) 
            this.addButton(mouseListener,buttons[i]);
    }
    /**
     * Creates a panel with given buttons, adding the given listener.
     * Creates the panel with the default layout, i.e. Y_AXIS.
     *
     * @param mouseListener a listener interested in clicks of the buttons
     * @param buttons a string array of button names
     */
    public IXButPanel(MouseListener mouseListener, Object[] buttons){
        this(mouseListener, BoxLayout.Y_AXIS, buttons);
    }
    
    /**
     * Adds a button with the given listener to this panel.
     * If the given object is a string, creates a new JButton with the string
     * as its label. If the button's actionCommand is empty, it is set to the
     * button's label.
     * Sets sensible minimum widths for the button and adds it to this panel.
     *
     * @param mouseListener a listener interested in clicks of the buttons
     * @param but a string or an AbstractButton that will be added
     */
    public void addButton(MouseListener mouseListener, Object but){
      AbstractButton button;
      if (but.getClass() == String.class) button = new JButton((String)but);
      else button = (AbstractButton) but;
      if (button.getActionCommand() == "") 
	button.setActionCommand(button.getText());
      java.awt.Dimension dim = button.getPreferredSize();
      double height = dim.getHeight();
      dim.setSize((double)IXGUIConsts.GOOD_BUT_WIDTH, height);
      button.setMinimumSize(dim);
      button.addMouseListener(mouseListener);
      this.add(button);
      this.buttons.add(button);
      if (groupFlag) group.add(button);
    }

  public String getSelectionCommand() {
    if (groupFlag) return group.getSelection().getActionCommand();
    else return "";		 
  }

  public void setButtonGroup(boolean group) {
    this.groupFlag = group;
    if (group) {
      if (this.group == null) this.group = new ButtonGroup();
      fillButtonGroup();
    }
    else this.group = null;
  }
  public boolean isButtonGroup() {
    return groupFlag;
  }

  public ButtonGroup getButtonGroup() {
    return group;
  }

  protected void fillButtonGroup() {
    Iterator butts = buttons.iterator();
    while (butts.hasNext())
      group.add((AbstractButton)butts.next());
  }

  public void setEnabled(boolean isEditable) {
    Iterator butts = buttons.iterator();
    while (butts.hasNext()) {
      AbstractButton butt = (AbstractButton)butts.next();
      butt.setEnabled(isEditable);
    }
  }
}
    
