/****************************************************************************
 * A simple panel with a label, a component, and optional buttons.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui;

import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.text.JTextComponent;
import javax.swing.table.*;
import java.awt.event.MouseListener;

import ix.util.Debug;
import ix.iface.ui.table.*;
    
/**
 * A simple panel with a label (title), a component, and optional buttons.
 * The given label will be at the top, the given component at the left and
 * buttons with given labels on the right.
 * Adds the given listener to the buttons and the list.
 */
public class IXEdPanel extends JPanel {
  public boolean isBordered = false;
  /** A sub-panel containing buttons */
  public IXButPanel butPanel;
  /** The panel's label (title) */
  public JLabel nameLabel;
  public TitledBorder border;
  /** The panel's item that may contain data. There are methods for setting
   * and getting data for JList and JTextComponent items
   */
  public JComponent itemComponent;
        
  /**
   * Creates a panel with a label (top left) or a border (depending on the
   * default)
   * Other components (a JComponent and a button panel) are usually added via
   * one of the other constructors; the label text is set in other
   * constructors.
   */
  public IXEdPanel(){
    // always make a label to the top left
    super(new GridBagLayout());
    makePanel();
  }
  /**
   * Creates a panel as above but with a given border setting.
   * If there is a border, the title of the panel will appear in the border,
   * otherwise it will appear in a label at the top left of the panel.
   * Other components (a JComponent and a button panel) are usually added via
   * one of the other constructors; the label text is set in other
   * constructors.
   */
  public IXEdPanel(boolean makeBorder) {
    // always make a label to the top left
    super(new GridBagLayout());
    isBordered = makeBorder;
    makePanel();
  }
  /**
   * Creates a panel with a title and a component (left).
   * If the border-flag is set, the title will go into the border, otherwise
   * a label will be used (top left).
   * If the component is a JList or a JTextArea, it is made scrollable
   * Also used by IXEdPanel constructors that have buttons.
   *
   * @param ml the mouse listener that is interested in component mouseEvents
   * @param makeBorder a flag determining whether the title is set in a border
   *   (true) or a label (false).
   * @param label a string that is used as the label (title of the panel).
   * @param component the component at the left of the panel
   */
  public IXEdPanel(MouseListener ml, boolean makeBorder, String label,
		   JComponent component){
    this(makeBorder);
    makePanelComponents(ml, label, component);
  }

  /**
   * Creates a panel with a title and a component (left) as above with default
   * border/label.
   *
   * @param ml the mouse listener that is interested in component mouseEvents
   * @param label a string that is used as the label (title of the panel).
   * @param component the component at the left of the panel
   */
  public IXEdPanel(MouseListener ml, String label, JComponent component){
    this();
    makePanelComponents(ml, label, component);
  }

  /**
   * Creates a panel with a label (top left) and buttons (right) only.
   *
   * @param ml the mouse listener that is interested in button mouseEvents
   * @param label a string that is used as the label (title of the panel).
   * @param buttons an array of strings that are button labels
   */
  public IXEdPanel(MouseListener ml, String label, String[] buttons){
    this();
    setLabel(label);
    this.butPanel = new IXButPanel(ml, buttons);
    GridBagConstraints gbc = 
      new GridBagConstraints(1,1,  //x,y
			     1,1,   //width,height
			     0.0,0.0,  //weight x,y
			     GridBagConstraints.EAST, //anchor
			     GridBagConstraints.NONE,  //fill
			     new Insets(0,2,0,2),0,0); //insets and padding
    this.add(this.butPanel, gbc);
  }
      
  /**
   * Creates a panel with a title, component (left) and buttons (right).
   * The title goes either into a label (top left) or into a border, depening
   * on the parameter.
   * Calls IXEdPanel(MouseListener ml, String label, JComponent component)
   * and makes the buttons itself.
   *
   * @param ml the mouse listener that is interested in component mouseEvents
   * @param makeBorder a flag determining whether the title is set in a border
   *   (true) or a label (false).
   * @param label a string that is used as the label (title of the panel).
   * @param component the component at the left of the panel
   * @param buttons an array of strings that are button labels
   */
  public IXEdPanel(MouseListener ml, boolean makeBorder, String label,
		   JComponent component, String[] buttons){
    this(ml, makeBorder, label, component);
    makeButtons(ml, buttons);
  }
  /**
   * Creates a panel as above but with the default border/label setting.
   *
   * @param ml the mouse listener that is interested in component mouseEvents
   * @param label a string that is used as the label (title of the panel).
   * @param component the component at the left of the panel
   * @param buttons an array of strings that are button labels
   */
  public IXEdPanel(MouseListener ml, String label, JComponent component, 
		   String[] buttons){
    this(ml, label, component);
    makeButtons(ml, buttons);
  }

  private void makePanel() {
    GridBagConstraints gbc = 
      new GridBagConstraints(0,0,  GridBagConstraints.REMAINDER,1, //x,y, w,h
			     0.0,0.0,  //weight x,y
			     GridBagConstraints.NORTHWEST, //anchor
			     GridBagConstraints.NONE,  //fill
			     new Insets(0,4,0,0),0,0); //insets and padding
    if (isBordered) {
      border = BorderFactory.createTitledBorder("some label");
      this.setBorder(border);
    }
    else {
      nameLabel = new JLabel("some label");
      this.add(nameLabel, gbc);
    }
  }

  /**
   * Puts the component into the panel (left) and sets the title.
   * If the border-flag is set, the title will go into the border, otherwise
   * a label will be used (top left).
   * If the component is a JList or a JTextArea, it is made scrollable
   * Also used by IXEdPanel constructors that have buttons.
   *
   * @param ml the mouse listener that is interested in component mouseEvents
   * @param label a string that is used as the label (title of the panel).
   * @param component the component at the left of the panel
   */
  private void makePanelComponents(MouseListener ml, String label, 
				   JComponent component){
    int fill = GridBagConstraints.BOTH;
    String c = "";
    if (component == null) c = "null"; else c = component.toString();
    //Debug.noteln("IXEdPanel: component is " + c);
    if (component.getClass() == JTextField.class) 
      fill = GridBagConstraints.HORIZONTAL;
    if (component.getClass() == JCheckBox.class) 
      fill = GridBagConstraints.NONE;
    GridBagConstraints gbc = 
      new GridBagConstraints(0,1, 1,1,  //x,y, width,height
			     2.0,1.0,  //weight x,y
			     GridBagConstraints.WEST, fill, //anchor, fill
			     new Insets(0,2,0,2),1,0); //insets and padding
    if ((component.getClass() == JList.class) || 
	(component.getClass() == JTextArea.class)) {
      // JLists and JTextAreas should be scrollable!
      JScrollPane jsp = new JScrollPane();
      jsp.getViewport().add(component);
      this.add(jsp, gbc);
    }
    else this.add(component, gbc);

    //strictly this bit should be obsolete with the above
    JComponent insideComponent = null;
    if (component.getClass() == JScrollPane.class) {
      //if the component is a scroll pane, the real component is the one inside
      try { //try to get the component inside
	insideComponent = (JComponent)
	  ((JScrollPane)component).getViewport().getComponent(0);
      }
      catch (ArrayIndexOutOfBoundsException e) {
      }
    }
    if (insideComponent == null) //no scroll pane or no inside component found
      itemComponent = component;
    else itemComponent = insideComponent;

    itemComponent.addMouseListener(ml);
    setLabel(label);
    this.setName(label);
  }
  private void makeButtons(MouseListener ml, String[] buttons){
    butPanel = new IXButPanel(ml, (Object[])buttons);
    GridBagConstraints gbc = 
      new GridBagConstraints(1,1,  //x,y
			     1,1,   //width,height
			     0.0,0.0,  //weight x,y
			     GridBagConstraints.EAST, //anchor
			     GridBagConstraints.NONE,  //fill
			     new Insets(0,2,0,2),0,0); //insets and padding
    this.add(this.butPanel, gbc);
  }

  public void setLabel(String label) {
    if (isBordered) border.setTitle(label);
    else nameLabel.setText(label);
  }
      
  /**
   * Gets the component of the panel.
   *
   * @return the JComponent that is at the left of the panel
   */
  public JComponent getItemComponent(){
    return this.itemComponent;
  }
  /**
   * Gets the current data of a JList component.
   *
   * @return an array of objects that are the current data model of the list
   */
  public Object[] getListData() {
    ListModel  model = ((JList)this.itemComponent).getModel();
    return getModelData(model);
  }
  /**
   * Gets the current data of a JComboBox component.
   *
   * @return an array of objects that are the current data model of the list
   */
  public Object[] getComboData() {
    ListModel  model = ((JComboBox)this.itemComponent).getModel();
    return getModelData(model);
  }

  /**
   * Gets the current data of a JTable component.
   *
   * @return an array of objects that are the current data model of the table
   */
  public Object[] getTableData() {
    TableModel  model = ((JTable)this.itemComponent).getModel();
    Debug.noteln("IXEP: Getting data from a table is not yet implemented");
    Object[] data = {};
    return data;
  }

  private Object[] getModelData(ListModel model) {
    int s = model.getSize();
    Object[] data = new Object[s]; 
    for (int i=0; i<s; i++) data[i] = model.getElementAt(i);
    return data;
  }

  /**
   * Gets the current data of a text component.
   *
   * @return a string that contains the current text of the component
   */
  public String getTextData(){
    return ((JTextComponent)this.itemComponent).getText().trim();
  }
  public Object getData(){
    if (JList.class.isAssignableFrom(this.itemComponent.getClass())) 
      return getListData();
    else if (JComboBox.class.isAssignableFrom(this.itemComponent.getClass()))
      return getComboData();
    else if (JTable.class.isAssignableFrom(this.itemComponent.getClass()))
      return getTableData();      
    else return getTextData();
  }
  public String getName(){
    return this.nameLabel.getText().trim();
  }

  /**
   * Sets the data of a JList component.
   *
   * @param data an array of objects that is to be used as the data of the list
   */
  public void setListData(Object[] data){
    this.itemComponent.setEnabled(false);
    if (JList.class.isAssignableFrom(this.itemComponent.getClass())) 
      ((JList)this.itemComponent).setListData(data);
    else if (JComboBox.class.isAssignableFrom(this.itemComponent.getClass())) {
      JComboBox box = (JComboBox)this.itemComponent;
      box.removeAllItems();
      for (int i=0; i<data.length; i++) box.addItem(data[i]);
    }
    else if (JTable.class.isAssignableFrom(this.itemComponent.getClass())) {
      TableModel model = ((JTable)itemComponent).getModel();
      if (AbstractIXTableModel.class.isInstance(model)) {
	((AbstractIXTableModel)model).setData(data);
      }
    }
    this.itemComponent.setEnabled(true);
  }
  /**
   * Sets selection of a JList or comboBox component.
   *
   * @param object an object that is to be selected in the list or comboBox
   */
  public void setSelectedItem(Object object){
    if (JList.class.isAssignableFrom(this.itemComponent.getClass()))
      ((JList)this.itemComponent).setSelectedValue(object, true);
    else if (JComboBox.class.isAssignableFrom(this.itemComponent.getClass()))
      ((JComboBox)this.itemComponent).setSelectedItem(object);
  }
  /**
   * Sets the current data of a text component.
   *
   * @param data a string that contains the text to put into the component
   */
  public void setTextData(Object data){
    //texts may sit in scrollPanes
    JTextComponent textComp;
    if (this.itemComponent.getClass() == JScrollPane.class) {
	JScrollPane pane = (JScrollPane) this.itemComponent;        
	textComp = (JTextComponent) pane.getViewport().getComponent(0);            
    } else { textComp = (JTextComponent) this.itemComponent;}
    textComp.setEnabled(false);
    if (data == null) textComp.setText("");    
    else textComp.setText(data.toString());    
    textComp.setEnabled(true);
  }

  public void setData(Object data) {
    this.itemComponent.setEnabled(false);
    //Debug.noteln("Setting data for " + nameLabel.getText());
    if (isListItem()) {
      //Debug.noteln(" data: " + data.toString());
      //Debug.noteln(" data class: " + data.getClass().getName());
      setListData((Object[])data);
    }
    else setTextData(data);
    this.itemComponent.setEnabled(true);
  }

  public void removeData() {
    this.itemComponent.setEnabled(false);
    //Debug.noteln("IXEdPanel: removing data for " + nameLabel.getText());
    if (isListItem()) {
      //Debug.noteln(" data: " + data.toString());
      //Debug.noteln(" data class: " + data.getClass().getName());
      String[] data = {};
      setListData(data);
    }
    else setTextData("");
    this.itemComponent.setEnabled(true);
  }

  public boolean isListItem() {
    Class itemClass = this.itemComponent.getClass();
    return ((JList.class.isAssignableFrom(itemClass))
	    || (JComboBox.class.isAssignableFrom(itemClass))
	    || (JTable.class.isAssignableFrom(itemClass)));
  }

  /*
  public void setData(Object[] data) {
    this.itemComponent.setEnabled(false);
    if (isListItem()) setListData(data);
    else setTextData(data);
    this.itemComponent.setEnabled(true);
  }
  */

  /**
   * Sets the renderer if the itemComponent is a JList or a JComboBox;
   * @return true if the renderer was set, false if not.
   */
  public boolean setRenderer(ListCellRenderer r) {
    //Debug.noteln("IXEdPanel: Set renderer in " + itemComponent.getClass());
    if (JList.class.isAssignableFrom(this.itemComponent.getClass())) {
      ((JList)itemComponent).setCellRenderer(r);
      return true;
    }
    else if (JComboBox.class.isAssignableFrom(this.itemComponent.getClass())) {
      ((JComboBox)itemComponent).setRenderer(r);
      return true;
    }
    /*
    else if (JTable.class.isAssignableFrom(this.itemComponent.getClass())) {
      ((JTable)itemComponent).setCellRenderer(r);
      return true;
    }
    */
    else return false;
  }
}
