/****************************************************************************
 * A simple panel with a label, a collection component, and optional buttons.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui;

import javax.swing.*;
import javax.swing.table.*;
import java.awt.event.*;

import ix.util.Debug;
import ix.iface.ui.table.*;
    
/**
 * A simple panel with a label (title), a component, and optional buttons.
 * The given label will be at the top, the given component at the left and
 * buttons with given labels on the right.
 * Adds the given listener to the buttons and the list.
 */
public class ListEditorPanel extends AbstractEditorPanel {

  public ListEditorPanel() { super();}
  public ListEditorPanel(MouseListener ml, String label, 
			 JComponent component) { 
    super(ml, label, component);
  }
  public ListEditorPanel(MouseListener ml, String label, String[] buttons) { 
    super(ml, label, buttons);
  }
  public ListEditorPanel(MouseListener ml, String label, 
			 JComponent component, String[] buttons) { 
    super(ml, label, component, buttons);
  }

  public boolean isListItem() {return true;}

  /**
   * Gets the current data of a collection component.
   * @return an object that is an array of objects taken from the component.
   */
  public Object getData() {
    Class iClass = itemComponent.getClass();
    if (JList.class.isAssignableFrom(iClass)) 
      return getListData();
    else if (JComboBox.class.isAssignableFrom(iClass)) 
      return getComboData();
    else if (JTable.class.isAssignableFrom(iClass)) 
      return getTableData();      
    else {
      Debug.noteln("Cannot get data from object of type " + iClass);
      return null;
    }
  }

  /**
   * Gets the current data of a JList component.
   * @return an array of objects that are the current data model of the list
   */
  private Object[] getListData() {
    ListModel  model = ((JList)this.itemComponent).getModel();
    return getModelData(model);
  }
  /**
   * Gets the current data of a JComboBox component.
   * @return an array of objects that are the current data model of the list
   */
  private Object[] getComboData() {
    ListModel  model = ((JComboBox)this.itemComponent).getModel();
    return getModelData(model);
  }
  /**
   * Gets the current data of a JTable component.
   * @return an array of objects that are the current data model of the table
   */
  private Object[] getTableData() {
    TableModel  model = ((JTable)this.itemComponent).getModel();
    Debug.noteln("LEP: Getting data from a table is not yet implemented");
    Object[] data = {};
    return data;
  }

  private Object[] getModelData(ListModel model) {
    int s = model.getSize();
    Object[] data = new Object[s]; 
    for (int i=0; i<s; i++) data[i] = model.getElementAt(i);
    return data;
  }

  public void setListData(Object[] data) {
    this.itemComponent.setEnabled(false);
    if (JList.class.isInstance(itemComponent)) {
      ((JList)this.itemComponent).setListData(data);
    }
    else if (JComboBox.class.isInstance(itemComponent)) {
      JComboBox box = (JComboBox)this.itemComponent;
      box.removeAllItems();
      for (int i=0; i<data.length; i++) box.addItem(data[i]);
    }
    else if (JTable.class.isInstance(itemComponent)) {
      TableModel model = ((JTable)itemComponent).getModel();
      if (AbstractIXTableModel.class.isInstance(model)) {
	((AbstractIXTableModel)model).setData(data);
      }
    }
    this.itemComponent.setEnabled(true);    
  }

  /**
   * Sets the data of a collection component.
   * @param data an array of objects that is to be used as the data 
   */
  public void setData(Object data) {
    setListData((Object[])data);
  }

  public void removeData() {
    String[] data = {};
    setListData(data);
  }

  public boolean setRenderer(ListCellRenderer r) {
    Class iClass = itemComponent.getClass();
    //Debug.noteln("ListEditorPanel: Set renderer in " + iClass);
    if (JList.class.isAssignableFrom(iClass)) {
      ((JList)itemComponent).setCellRenderer(r);
      return true;
    }
    else if (JComboBox.class.isAssignableFrom(iClass)) {
      ((JComboBox)itemComponent).setRenderer(r);
      return true;
    }
    /*
    else if (JTable.class.isAssignableFrom(iClass)) {
      ((JTable)itemComponent).setCellRenderer(r);
      return true;
    }
    */
    else return false;
  }

}
