/****************************************************************************
 * An abstract class for table models.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui.table;

import ix.iface.ui.*;
import ix.iface.ui.util.*;
import ix.util.*;

import java.util.*;
import java.lang.reflect.*;
import javax.swing.table.*;

/****************************************************************************
 * An abstract class for table models.
 *
 * Keeps track of the base objects that are shown in the table.
 * Implements most functions required by JTable and IXTable objects to work.
 * 
 * Implement the following functions (see documentation below):
 * <ul>
 * <li> protected Object getCellValueAt(Object baseObject, int columnIndex)
 * <li> public String[] getColumnNames();
 * </ul>
 *****************************************************************************
 */
public abstract class AbstractIXTableModel extends AbstractTableModel {
  protected ArrayList columnFields = new ArrayList();
  protected ArrayList columnNames = new ArrayList();
  protected ArrayList columnClasses = new ArrayList();
  protected Class objectClass;


  /** a list of objects that appear in the table */
  protected List rows = new LinkedList();

  
  /**
   * Makes an empty model.
   */
  public AbstractIXTableModel() {super();}

  /**
   * Makes a model that contains the given nodes.
   */
  public AbstractIXTableModel(Object[] theRows) {
    super();
    setData(theRows);
  }

  /**
   * Makes a model as above but with the given strings as column names.
   */
  public AbstractIXTableModel(Object[] theRows, String[] columns) {
    this(theRows);
    setColumnNames(columns);
  }

  /**
   * Makes a model that contains the given nodes and derives which fields
   * to use. 
   */
  public AbstractIXTableModel(Class theClass, Object[] theRows) {
    this(theRows);
    objectClass = theClass;
    deriveColumnFields();
  }

  /**
   * Sets the table's rows to the given data objects and notifies listeners.
   */
  public void setData(Object[] data){
    rows.clear();
    for (int i = 0; i < data.length; i++) {
      rows.add(data[i]);
    }
    fireTableDataChanged();
  } 

  /*
  private Object findData(Object data) {

  }
  */
          
  /**
   * Adds the given data object to the table's rows if it is not already there
   * and notifies listeners.
   */
  public void addData(Object data) {
    if (!rows.contains(data)) {
      rows.add(data);
      fireTableDataChanged();
    }
  }
  /**
   * Removes the given data object from the table's rows and notifies
   * listeners if it worked.
   */
  public void removeData(Object data) {
    if (rows.remove(data)) fireTableDataChanged();
  }
  /**
   * Removes all data objects from the table's rows and notifies listeners.
   */
  public void clearData() {
    rows.clear();
    fireTableDataChanged();
  }

  /** 
   * Gets the object that is in the given row.
   * @return the object or null if there is no such row.
   */
  public Object getRowObject(int rowIndex) {
    if (rowIndex < 0) return null;
    else return rows.get(rowIndex);    
  }

  /** 
   * Gets the row index of the given object
   * @return the object's index or -1 if there is no such object.
   */
  public int getObjectRow(Object object) {
    return rows.indexOf(object);    
  }

  /**
   * Gets the value for a cell in the table defined by the given indices.
   * Looks up the object and calls the abstract object-index method.
   */ 
  public Object getValueAt(int rowIndex, int columnIndex) { 
    Object o = rows.get(rowIndex);    
    return getCellValueAt(o, columnIndex);
  }

  /** JTable required TableModel function. */
  public int getRowCount() {
    return rows.size();
  }
  /** JTable required TableModel function. */
  public int getColumnCount() {
    String[] names = getColumnNames();
    return names.length;
  }
  /** JTable required TableModel function. */
  public String getColumnName(int index) {
    String[] names = getColumnNames();
    return names[index];
  }
  /** JTable required TableModel function (default is Object). */
  public Class getColumnClass(int index) {
    Class colClass = null;
    try {
      colClass = (Class)columnClasses.get(index);
    }
    catch (IndexOutOfBoundsException e) {
      colClass = super.getColumnClass(index);
    }
    return colClass;
  }

  /**
   * Gets the object's value for the given column index.
   * @return an object that can be rendered by the table.
   */
  public abstract Object getCellValueAt(Object baseObject, int columnIndex);

  /**
   * Sets the column titles to be used in the table
   */
  public void setColumnNames(String[] fieldNames) {
    columnNames = new ArrayList(Arrays.asList(fieldNames));
  }
  /**
   * Sets the column classes of the table
   */
  public void setColumnClasses(Class[] fieldClasses) {
    columnClasses = new ArrayList(Arrays.asList(fieldClasses));
  }
  /**
   * Gets the strings to use as column names.
   */
  protected String[] getColumnNames() {
    String[] s = {};
    return (String[])columnNames.toArray(s);
  }
  /**
   * Sets the fields to be used as columns in the table
   */
  public void setColumnFields(Class objectClass, String[] fieldNames) {
    for (int i=0; i<fieldNames.length; i++) {
      String fieldName = fieldNames[i];
      try {
	Field field = objectClass.getField(fieldName);
	String title = UIUtil.toCapitalised(fieldName);
	Class fieldClass = field.getType();
	columnFields.add(fieldName);
	columnNames.add(title);
	columnClasses.add(fieldClass);
      }
      catch (Exception e) {
	Debug.noteln("AIXTM: Problem setting column field " + fieldName);
	Debug.noteException(e);
      }
    }
  }



  /*-------------------------------------------------------------------------
   *
   *        Useful methods for generic tables
   *
   *-------------------------------------------------------------------------
   */

  /**
   * Collects usable fields from the objectClass 
   */
  protected void deriveColumnFields() {
    if (objectClass == null) return;
    //class fields
    ArrayList fields = new ArrayList(Arrays.asList(objectClass.getFields()));
    Iterator e = fields.iterator();
    while (e.hasNext()) {
      Field field = (Field)e.next();
      if (useField(field)) {
	columnFields.add(field.getName());
	columnNames.add(UIUtil.toCapitalised(field.getName()));
      }
    }
  }

  /**
   * Checks whether the given field can be used.
   */
  private static boolean useField(Field field) {
    int mod = field.getModifiers();
    //do not display protected, private, static, or name fields
    return (!Modifier.isStatic(mod) && 
	    !Modifier.isPrivate(mod) &&
	    !Modifier.isProtected(mod));
  }


  /**
   * Sets the fields to be used as table columns
   */
  protected void defaultSetColumnFields(String[] fieldNames) {
    if (objectClass == null) return;
    columnFields.clear();
    if ((fieldNames == null) || (fieldNames.length == 0)) return;
    else {
      for (int i=0; i<fieldNames.length; i++) {
	String name = fieldNames[i];
	try {
	  objectClass.getField(name);
	  columnFields.add(name);
	  columnNames.add(UIUtil.toCapitalised(name));
	}
	catch (Exception e) {
	  Debug.noteException(e);
	  try {
	    String lName = UIUtil.toUncapitalised(name);
	    objectClass.getField(lName);
	    columnFields.add(lName);
	    columnNames.add(name);
	  }
	  catch (Exception ee) {
	    Debug.noteException(ee);
	  }
	}
      }
    }
  }

  /**
   * Looks up the value in the columnField of the given index.
   * @return the value object or null if something goes wrong.
   */
  protected Object defaultGetCellValueAt(Object o, int columnIndex) {
    if (objectClass == null) return null;
    if ((columnFields == null) || columnFields.isEmpty()) return null;
    try {
      String fieldName = (String)columnFields.get(columnIndex);
      Field field = objectClass.getField(fieldName);
      return field.get(o);
    }
    catch (Exception e) {
      Debug.note("Exception during defaultGetCellValueAt(" + o.toString() + 
		 ", " + columnIndex + ")");
      Debug.noteException(e);
      return null;
    }
  }



}
