/****************************************************************************
 * An abstract class for tree table models.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui.table;

import ix.iface.ui.*;
import ix.iface.ui.util.*;
import ix.util.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.*;

/****************************************************************************
 * An abstract class for tree table models.
 *
 * Requires data objects to be TreeTable objects. 
 * Creates IXTreeTableNode objects to wrap around the data objects. These 
 * IXTreeTableNode objects hold id information (base-object and parent line)
 * and expansion information (indentation of the node and whether it is 
 * expanded or not).
 *
 * By default, the children of nodes are not displayed
 * When clicking on the node, the following happens:
 * <ul>
 * <li> if the node is a leaf: nothing
 * <li> if the node has children and they are displayed: un-display them
 * <li> if the node has children and they are NOT displayed: display them
 * </ul>
 * Multiple parents are no problem because by default nodes children are not 
 * displayed. When a child is displayed more than once, it has a different 
 * parent-line and different IXTreeTableNode obejcts are used. <p>
 *
 * As an example, see ix.iview.table.ActionTreeTableModel
 *****************************************************************************
 */
public abstract class AbstractTreeTableModel extends AbstractIXTableModel 
implements TableMouseListener {

  /** the indentation steps used to indent children */
  public static final int INDENT_STEP = 4;
  /** the character used to indicate a leaf (no children) */
  public static final char LEAFC = ' ';
  /** the character used to indicate an expanded node */
  public static final char OPENC = '-';
  /** the character used to indicate an un-expanded node */
  public static final char CLOSEDC = '+';

  public static final String imageDir = "images/";
  public static final ImageIcon LEAFI = null;
  public static final ImageIcon OPENI = 
     Util.resourceImageIcon("ix-symbol-fold.gif");
  public static final ImageIcon CLOSEDI = 
     Util.resourceImageIcon("ix-symbol-unfold.gif");

  /** a collection of IXTreeTableNodes that have been generated */
  protected ArrayList nodes = new ArrayList();
  
  /**
   * Makes an empty model.
   */
  public AbstractTreeTableModel(){
    super();
  }
  /**
   * Makes a model that contains the given nodes at top level.
   */
  public AbstractTreeTableModel(TreeNode[] theNodes){
    super();
    //add the node, set indentations of nodes to 0
    setData(theNodes);
  }

  /**
   * Sets the table's rows to the given TreeNode objects and notifies
   * listeners.
   * Creates a top-level IXTreeTableNode object for each data object.
   * Overwrites the super method.
   */
  public void setData(Collection theNodes){
    //Debug.noteln("AbstractTTM: setting table data");
    clearData();
    for (Iterator i = theNodes.iterator(); i.hasNext(); ) {
      Object node = i.next();
      TreeNode eNode = makeIXTreeTableNode(node);
      if (eNode == null) {
	Debug.noteln("WARNING: Tree table is ignoring node", node.toString());
	Debug.noteln(" It is of class " + node.getClass().getName());
	Debug.noteln(" and not a valid TreeNode.");
      }
      else {
	if (!rows.contains(eNode)) {
	  rows.add(eNode);
	  nodes.add(eNode);
	}
      }
    }
    fireTableDataChanged();
  }    
  /**
   * Sets the table's rows to the given TreeNode objects and notifies
   * listeners.
   * Creates a top-level IXTreeTableNode object for each data object.
   * Overwrites the super method.
   */
  public void setData(Object[] theNodes){
    setData(Arrays.asList(theNodes));
  }  

  public void clearData() {
    rows.clear();
    nodes.clear();
    nodeMap.clear();
  }
  public void clearTable() {
    rows.clear();
    nodes.clear();
    fireTableDataChanged();
  }
  

  /**
   * Adds the given TreeNode object to the table's rows and notifies listeners.
   * Creates a top-level IXTreeTableNode object for the data object.
   * Overwrites the super method.
   */
  public void addData(Object data) {
    //Debug.noteln("AbstractTTM: adding table data");
    if (data == null) {
      Debug.noteln("AbstractTTM: Cannot add null data to table");
      (new Throwable()).printStackTrace();
    }
    else {
      TreeNode eNode =  makeIXTreeTableNode(data);
      if (!rows.contains(eNode)) {
	rows.add(eNode);
	nodes.add(eNode);
	fireTableDataChanged();
      }
    }
  }
  /**
   * Removes the given TreeNode object from the table's rows and notifies
   * listeners.
   * Creates a top-level IXTreeTableNode object for the data object.
   * Overwrites the super method.
   */
  public void removeData(Object data) {
    //Debug.noteln("AbstractTTM: adding table data");
    TreeNode eNode = getTreeNode(data);
    if (rows.contains(eNode)) {
      rows.remove(eNode);
      nodes.remove(eNode);
      fireTableDataChanged();
    }
  }
          
  /**
   * Adds a node to the model in the given place with the given indentation.
   * Also notifies listeners. This happens when nodes are expanded.
   */
  private void addNode(IXTreeTableNode eNode, int place){
    rows.add(place, eNode);  //make a new object
    fireTableDataChanged();
  }
  
  /**
   * Removes the given number of nodes from the given index place in the table.
   * Also notifies listeners. This happens when nodes are collapsed.
   */
  public void removeNodes(int number, int place){
    for (int i = 0; i < number; i++) 
      rows.remove(place);
    fireTableDataChanged();
  }

  /**
   * Sets the column titles to be used in the table
   */
  public void setColumnNames(String[] fieldNames) {
    super.setColumnNames(fieldNames);
    columnNames.add(0,"");
    Class tClass = determineTreeTypeClass();
    columnClasses.add(0, tClass);
  }
  /**
   * Sets the fields to be used as columns in the table
   */
  public void setColumnFields(Class objectClass, String[] fieldNames) {
    super.setColumnFields(objectClass, fieldNames);
    columnFields.add(0,"treeType");
    columnNames.add(0,"");
    Class tClass = determineTreeTypeClass();
    columnClasses.add(0, tClass);
  }

  private Class determineTreeTypeClass() {
    if (OPENI != null) return ImageIcon.class;
    else return Character.class;
    //return ImageIcon.class;
  }


  public Object getRowObject(int rowIndex) {
    if (rowIndex < 0) return null;
    IXTreeTableNode eNode = (IXTreeTableNode)rows.get(rowIndex); 
    return eNode.node;
  }
  public Object getRowNode(int rowIndex) {
    return rows.get(rowIndex); 
  }
  public int getObjectRow(Object object) {
    //find the node's IXTreeTableNode, return that node's index
    TreeNode iNode = getTreeNode(object);
    return super.getObjectRow(iNode);
  }


  /**
   * Get the value for a cell in the table defined by the parameters.
   * only the node name is different from normal node tables, so the other
   * cells can be done by any old table model.
   */ 
  public Object getValueAt(int rowIndex, int columnIndex) { 
    IXTreeTableNode eNode = (IXTreeTableNode)rows.get(rowIndex);    
    if (columnIndex == 0) {
      return getTreeNodeType(eNode);
    } 
    if (columnIndex == 1) {
      return getTreeValueAt(eNode);
    } 
    else return getCellValueAt(eNode, columnIndex-1);
  }

  /**
   * Gets the value that indicates the tree-position (indented name).
   * @return an object that can be rendered by the table.
   */
  protected Object getTreeValueAt(IXTreeTableNode eNode) { 
    return toIndentedString(eNode);
  }

  /**
   * Gathers the node's children and splices them into the table.
   */
  public void expandNode(IXTreeTableNode eNode, int rowIndex){
    //get the node's children, set their indentation, and
    // splice them into the nodes list
    if (eNode.expanded) return;
    Enumeration kids = eNode.children();
    if (kids != null) {  
      //get the node parentLine to identify it,
      //get its IXTreeTableNode object, lookup the indentations
      //add a bit of indentation for children, add node to parents line
      //add the children to the rows just after, make children IXTreeTableNodes
      IXTreeTableNode eKid;
      LinkedList parents = new LinkedList(eNode.parentLine);
      parents.addLast(eNode.node);
      int indent = eNode.indent + INDENT_STEP;
      while (kids.hasMoreElements()) {
	Object kid = kids.nextElement();
	//Debug.noteln("ATreeTabMod: working on kid", kid.toString());
	eKid = (IXTreeTableNode)getTreeNode(kid, parents, indent);
	//add the kid to the table
	addNode(eKid, rowIndex+1);
	eNode.kids.add(eKid);
      }
      //remember this node is expanded
      eNode.setExpanded(true);
    }
  }
       
  
  /**
   * Removes the node's children from the table and notes it as unexpanded.
   */
  public void unexpandNode(IXTreeTableNode eNode, int index) {
    if (!eNode.expanded) return;
    int i = eNode.kids.size();
    if (i != 0) {
     // remove the right number of children just after the selected one
      removeNodes(i, index+1);
      eNode.kids.clear();
    }
    eNode.setExpanded(false);
  }

  /**
   * Produces a string that indicates the level in the tree, including initial
   * symbol.
   */
  private String toIndentedString(IXTreeTableNode eNode){
    String theValue = new String();
    //theValue = theValue + getExpansionChar(eNode) + " ";
    for (int i=0; i < eNode.indent; i++) 
      theValue = theValue + " ";
    theValue = theValue + getNameString(eNode);
    return theValue;
  }

  /**
   * Gets a character or icon to use as a reflection of the node's tree-status.
   * E.g. + for collapsed with children, - for expanded with children,
   * empty for no children.
   */
  public Object getTreeNodeType(IXTreeTableNode eNode) {
    if (ImageIcon.class.isAssignableFrom(getColumnClass(0))) 
      return getTypeIcon(eNode);
    else 
      return getTypeChar(eNode);
  }

  private ImageIcon getTypeIcon(IXTreeTableNode eNode) {
    ImageIcon icon = null;
    if (eNode.isLeaf()) icon = LEAFI;
    else if (eNode.expanded) icon = OPENI;
    else icon = CLOSEDI;
    return icon;
  }
  private Character getTypeChar(IXTreeTableNode eNode) {
    char e;
    if (eNode.isLeaf()) e = LEAFC;
    else if (eNode.expanded) e = OPENC;
    else e = CLOSEDC;
    return new Character(e);
  }

  /**
   * If the left button was clicked on the tree-column, expand/collapse node.
   */
  public void cellClicked(MouseEvent me, int rowIndex, int columnIndex) {
    //Debug.noteln("AbTTM: cell clicked");
    int mod = me.getModifiers();
    //only interested in left button on tree column
    if (SwingUtilities.isLeftMouseButton(me)
	&& (columnIndex == 0)) {
      IXTreeTableNode eNode = (IXTreeTableNode) this.rows.get(rowIndex);
      if (!eNode.isLeaf()) {
	if (eNode.expanded) unexpandNode(eNode, rowIndex);
	else expandNode(eNode, rowIndex);
      }
    }
  }

  /**
   * Gets the object's value for the given column index.
   * @return an object that can be rendered by the table.
   */
  public abstract Object getCellValueAt(Object node, int columnIndex);
  /**
   * Gets the string to use for the tree-column (the name of the node).
   */
  public abstract String getNameString(TreeNode node);



  /*-------------------------------------------------------------------------
   *
   *        Useful events for listeners
   *
   *-------------------------------------------------------------------------
   */
  /**
   * The given object changed - find its tree node, fire that and redo children
   */
  public void fireObjectChanged(Object o) {
    //Debug.noteln("AbTTM: got object change for", o);
    IXTreeTableNode node = (IXTreeTableNode)getTreeNode(o);
    //Debug.noteln("  got node", node);    
    int row = super.getObjectRow(node);
    //Debug.noteln("  got row", row);    
    if (node.expanded) {
      //Debug.noteln("  sorting expanded node");    
      //unexpand, expand
      unexpandNode(node, row);
      expandNode(node, row);
    }
    fireTableDataChanged(); //expand/unexpand will fire this too
  }


  /*-------------------------------------------------------------------------
   *
   *        Useful methods for generic tables
   *
   *-------------------------------------------------------------------------
   */

  /**
   * Looks up the value in the first columnField.
   * @return the value object or null if something goes wrong.
   */
  public String defaultGetNameString(TreeNode node) {
    return (String)super.defaultGetCellValueAt(node, 0);
  }
  

  /**
   * Collects usable fields from the objectClass 
   */
  protected void deriveColumnFields() {
    super.deriveColumnFields();
    //now find a good field to use as a name field
    int nameIndex = findNameField(columnFields);
    String nameField = (String)columnFields.get(nameIndex);
    columnFields.remove(nameIndex);
    columnFields.add(0, nameField);
    String n = (String)columnNames.get(nameIndex);
    columnNames.remove(nameIndex);
    columnNames.add(0, nameField);
  }

  /**
   * Looks for a field that sounds like a good name field and returns its
   * index in the given list.
   */
  private static int findNameField(ArrayList fieldNames) {
    int index = fieldNames.indexOf("name");
    if (index != -1) return index;
    else {
      index = fieldNames.indexOf("Name");
      if (index != -1) return index;
      else {
	String fName;
	for (int i = 0; i<fieldNames.size(); i++) {
	  fName = UIUtil.toCapitalised((String)fieldNames.get(i));
	  if (fName.indexOf("NAME") != -1)
	    return i;
	}
      }
    }
    return 0;  //no useful field found, so use the first one.
  }



  //--------------------IXTreeNode vs node object--------------------------

  /**
   * A map for looking up a TreeNode for a Object.
   * Use: put(Object, TreeNode) 
   */
  public HashMap nodeMap = new HashMap();

  /**
   * Get the (possibly new) TreeNode that is associlated with
   * the given Object.
   */
  public TreeNode getTreeNode(Object object) {
    if (object == null) return null;

    TreeNode tNode = (TreeNode)nodeMap.get(object);
    if (tNode == null) {
      tNode = makeIXTreeTableNode(object);
      addTreeNode(object, tNode);
    }
    return tNode;
  }
  public TreeNode getTreeNode(Object object, List parents, int indentation) {
    if (object == null) return null;

    //make a node to look up
    IXTreeTableNode node = makeIXTreeTableNode(object, parents, indentation);
    int n = nodes.indexOf(node);
    //if we already had this node, use the old one. else remember new one
    if (n != -1) node = (IXTreeTableNode)nodes.get(n);
    else nodes.add(node);  
    return (TreeNode)node;
  }

  /** Associate the given TreeNode to the given AgendaItem */
  public void addTreeNode(Object object, TreeNode tNode) {
    nodeMap.put(object, tNode);
  }

  public abstract IXTreeTableNode makeIXTreeTableNode(Object node);
  /**
   * indentation is number of blanks to put in front of node (n*INDENT_STEP)
   */
  public IXTreeTableNode makeIXTreeTableNode(Object node, List parents,
					     int indentation) {
    IXTreeTableNode iNode = makeIXTreeTableNode(node);
    iNode.setParents(parents);
    iNode.indent = indentation;
    return iNode;
  }


}
