/****************************************************************************
 * Abstract table that listens to the mouse, has values with up to 4D and
 * sensible column widths.
 * 
 * @author Jussi Stader
 * @version 1.0
 ****************************************************************************
 */
package ix.iface.ui.table;


import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import ix.iface.ui.*;
import ix.iface.ui.*;
import ix.util.*;

/****************************************************************************
 * Abstract table that listens to the mouse, has values with up to 4D and
 * sensible column widths.
 *
 * The IXTable is used to provide a standard table behaviour for displaying
 * up to fourD displays of fields (as provided by the NDRenderer), adjusted
 * column widths, no re-ordering of columns, adding the model as a
 * TableModelListener if it is one.<p>
 * Can be used for normal tables or tree tables. <p>
 *
 * Example code for making a TreeTable happen:
 * <PRE><code>
 *   ...
 *     TreeNode[] tna = {};
 *     TreeNode[] nodes = domain.getAllRefinements().toArray(tna);
 *     ActionTreeTableModel tableModel = new ActionTreeTableModel(nodes);
 *     IXTable table = new IXTable(tableModel);
 *     table.setPreferredScrollableViewportSize(new Dimension(200, 100));
 *     JScrollPane jsp = new JScrollPane();
 *     jsp.getViewport().add(table);
 *     panel.add(jsp);
 *   ...
 * </code></PRE>
 ****************************************************************************
 */

public class IXTable extends JTable implements MouseListener {

  /** The TableModelListeners that have been added */
  private ArrayList tmListeners = new ArrayList();
  
  /** 
   * Makes a single-selection table that listens to mouse, shows grid,
   * renders 4D.
   */
  public IXTable(){
    //    Insets ins = getInsets();
    //    setSize(ins.left + ins.right + 430,ins.top + ins.bottom + 270);
    
    setShowGrid(true);
    this.addMouseListener(this);
    this.getSelectionModel()
      .setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    
    NDRenderer ndr = new NDRenderer(false);
    ndr.setToolTipText("Click for choices");
    IconRenderer ir = new IconRenderer(false);
    ir.setToolTipText("click to expand/collapse");
    this.setDefaultRenderer(FourField.class, ndr);
    this.setDefaultRenderer(Icon.class, ir);
    //by default columns are not reorderable
    this.getTableHeader().setReorderingAllowed(false);
  }

  /** 
   * Makes a table and sets the given model, adjusts column widths, registers
   * model as mouse listener (if appropriate).
   */
  public IXTable(AbstractTableModel model) {
    this();
    this.setModel(model);
    if (TableMouseListener.class.isAssignableFrom(model.getClass())) 
      this.addTableMouseListener((TableMouseListener)model);
    //this.getTableHeader().setReorderingAllowed(false);
  }

  public void setModel(TableModel model) {
    super.setModel(model);
    adjustWidths();
  }


  /**
   * Adjusts column widths to sensible values.
   * String: 100, FourField: 100, boolean: 25, others 50.
   * Relative width is preserved on resizing.
   */
  protected void adjustWidths() {
    TableColumn column;
    Class colClass;
    TableModel model = getModel();
    for (int i = 0; i < model.getColumnCount(); i++) {
      column = getColumnModel().getColumn(i);   //work on column i
      colClass = model.getColumnClass(i);   //get class for width
      //Debug.noteln("IXTable: doing column class " + colClass);
      if ((String.class.isAssignableFrom(colClass)) ||
	  (FourField.class.isAssignableFrom(colClass)))
	column.setPreferredWidth(100);
      else if (Boolean.class.isAssignableFrom(colClass)) 
	column.setPreferredWidth(15);
      else if ((Icon.class.isAssignableFrom(colClass)) ||
	       (ImageIcon.class.isAssignableFrom(colClass))) {
	column.setPreferredWidth(15);
	column.setMinWidth(15);
	column.setMaxWidth(15);
	column.setResizable(false);
      }
      else if (Character.class.isAssignableFrom(colClass))
	column.setPreferredWidth(15);
      else column.setPreferredWidth(50);
      //Debug.noteln(" set width to "+ column.getPreferredWidth());
    }
  }

  public void addTableMouseListener(TableMouseListener tml) {
    tmListeners.add(tml);
  }
  public void removeTableMouseListener(TableMouseListener tml) {
    while (tmListeners.contains(tml)) {
      tmListeners.remove(tmListeners.indexOf(tml));
    }
  }

  /** 
   * Send a cellClicked notification to all registered TableMouseListeners.
   * Passes on the mouse event and the row/column index that has been clicked.
   * Works for single-selection tables only.
   */
  public void mouseClicked(MouseEvent me) {
    //Debug.noteln("IXTable - mouse clicked");
    int row = getSelectedRow();
    int col = getSelectedColumn();
    //Debug.noteln(" - row: " + row + " column: " + col);
    Iterator e = tmListeners.iterator();
    while (e.hasNext()) 
      ((TableMouseListener)e.next()).cellClicked(me,row,col);
  }
  public void mouseEntered(MouseEvent me) {}
  public void mouseExited(MouseEvent me) {}
  public void mousePressed(MouseEvent me) {}
  public void mouseReleased(MouseEvent me) {}


}
