/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sun Jun  9 19:44:30 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.iface.util;

import javax.swing.JOptionPane;
import java.awt.event.*;
import java.util.*;

import ix.util.*;

/**
 * A wrapper to put around an ActionListener to catch any exceptions
 * or errors the listener throws and report them to the user.
 */
public class CatchingActionListener implements ActionListener {

    ActionListener innerListener;

    public CatchingActionListener(ActionListener innerListener) {
	this.innerListener = innerListener;
    }

    /**
     * Calls the inner listener's ActionPerformed method on the same
     * event, but inside a <code>try-catch</code> statement that
     * catches any exception or error thrown.  The ActionEvent and the
     * exception or error are then passed to this object's handleException
     * method after a call to <code>Debug.noteException(e)</code>.
     *
     * @see ix.util.Debug#noteException(Throwable)
     */
    public void actionPerformed(ActionEvent event) {
	try {
	    innerListener.actionPerformed(event);
	}
	catch (Throwable except) {
	    Debug.noteException(except);
	    handleException(event, except);
	}
    }

    /**
     * Called if the inner listener's ActionPerformed method throws
     * an exception or error.  This method puts up an error message dialog
     * that displays the ActionEvent's action command and a description
     * or the exception or error.
     */
    public void handleException(ActionEvent event, Throwable except) {
	String command = event.getActionCommand();
	JOptionPane.showMessageDialog(null,
	    new Object[] {"Problem during command " + Strings.quote(command),
			  description(except)},
	    "Error during " + command,
	    JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Returns a description suitable for display in a dialog.
     * Because an error or exception's message is not always sufficiently
     * meaningful on its own, it is prefixed by the class name (minus
     * the package name).  Long messages are split into lines.
     *
     * @see ix.util.Strings#foldLongLine(String)
     */
    public static String description(Throwable t) {
	return Strings.foldLongLine
	    (Strings.afterLast(".", t.getClass().getName()) + ": " +
	     t.getMessage());
    }

    /*
     * Factory.
     */

    static Map listenerCache = new WeakHashMap();

    /**
     * Returns the outer listener, if any, that was recorded for the
     * specified inner listener, or else constructs and records a
     * new CatchingActionListener.  The listeners are recorded in
     a WeakHashMap.
     */
    public static ActionListener listener(ActionListener inner) {
	ActionListener outer = (ActionListener)listenerCache.get(inner);
	if (outer == null) {
	    outer = new CatchingActionListener(inner);
	    listenerCache.put(inner, outer);
	}
	return outer;
    }

    /**
     * Records a specific outer listener for cases where it must
     * be an instance of class other than CatchingActionListener,
     * for instance a subclass.
     */
    public static void recordListener(ActionListener inner,
				      ActionListener outer) {
	listenerCache.put(inner, outer);
    }

}

// Issues:
// * Should we display the class of the error or exception?  Some messages
//   don't make sense without it, but some class names will look much too
//   "implementational", e.g. ix.util.lisp.LispReadException.
