/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed May  8 02:29:35 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.icore.*;
import ix.icore.domain.Refinement;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

/**
 * An agenda of activities
 */
public class ActivityAgenda extends Agenda {

    public ActivityAgenda(PanelController controller) {
	super(controller);
    }

    public AgendaItem makeItem(LList pattern) {
	return new ActivityItem(new Activity(pattern));
    }

    public void addItem(AgendaItem item) {
	item.setPattern(modelManager.putVariablesInPattern(item.getPattern()));
	modelManager.addNode(item);
	item.computeStatus();
	super.addItem(item);
    }

    public void addHandlerActions(AgendaItem item) {
	super.addHandlerActions(item);

	// Expansion and manual primitive options from schemas
	// in the domain.
	// /\/: Note that we can no longer use getMatchingSchemas(),
	// because we don't want the schemas to be instantiated.
	for (Iterator i = controller.domain.getRefinements().iterator();
	     i.hasNext();) {
	    Refinement r = (Refinement)i.next();
	    MatchEnv e = Matcher.match(r.getPattern(), item.getPattern());
	    if (e == null) {
		continue;
	    }
	    else if (Collect.isEmpty(r.getNodes())
		     && Collect.isEmpty(r.getConstraints())) {
		// When there are no subnodes or constraints, the
		// refinement name still specifies a way of doing
		// it manually.
		item.addAction(new HandlerAction.Manual(r.getName()));
	    }
	    else {
		// Add action to expand using the refinement.
		item.addAction(new ExpandAction(item, r, e));
	    }
	}
    }

    public void refinementAdded(Refinement r) {
	// See if any activities now have a new way to expand.
	for (Iterator i = items.iterator(); i.hasNext();) {
	    AgendaItem item = (AgendaItem)i.next();
	    MatchEnv env = Matcher.match(r.getPattern(), item.getPattern());
	    if (env == null) {
		continue;
	    }
	    else if (Collect.isEmpty(r.getNodes())
		     && Collect.isEmpty(r.getConstraints())) {
		Debug.noteln("New manual primitive for", item);
		item.addAction(new HandlerAction.Manual(r.getName()));
	    }
	    else {
		Debug.noteln("New expansion for", item);
		item.addAction(new ExpandAction(item, r, env));
	    }
	}
    }

    public void expandItem(AgendaItem item, Refinement instructions) {
	// This is really a kind of issue handler.  /\/
	// Avoid null pointer exceptions in debugging code:  /\/
	Refinement r = instructions;
	if (r.getName() == null)
	    r.setName("temporary partial schema");
	// Note that we don't add the action to the issue.  /\/
 	// Instead we handle "directly".
	HandlerAction act = new ManualExpandAction(item, r);
	item.addAction(act);	// ??? but we do add it ??? /\/
	act.handle(item);
	fireItemHandled(item, act);
    }

    /*
     * HandlerAction classes
     */

    /**
     * Handles an issue by expanding it into subissues
     * as specified by a schema.
     */
    class ExpandAction extends HandlerAction {

	AgendaItem item;
	Refinement refinement;
	MatchEnv env;

	List filterConds;
	MatchEnv filterEnv = null;

	ExpandAction(AgendaItem item, Refinement refinement, MatchEnv env) {
	    this.item = item;
	    this.refinement = refinement;
	    this.env = env;
	    this.filterConds = refinement.getFilterConditions();
	    this.shortDescription = "Expand using " + refinement.getName();
	}

	public boolean isStillValid() {
	    env = Matcher.match(refinement.getPattern(), item.getPattern());
	    return env != null;
	}

	public boolean isReady() {
	    return filterConds.isEmpty() || filterEnv != null;
	}

	public void computeStatus() {
	    if (!filterConds.isEmpty())
		filterEnv = modelManager.evalFilters(filterConds, env);
	}

	public void handle(AgendaItem item) {
	    Debug.assert(isReady(),
	        this + " asked to handle " + item + " when not ready");

	    // That an option's been selecte implies that execution
	    // has begun.  /\/: But subnodes must be put in first
	    // or they will get the wrong status as they are added
	    // (because parent will be executing and orderings won't
	    // yet be in so it will look like all preconditions have
	    // been met).
	    Debug.assert(item == this.item);
	    do_expansion(item);
	    Debug.assert(item.getStatus() == Status.POSSIBLE);
	    item.setStatus(Collect.isEmpty(item.getChildren())
			   ? Status.COMPLETE : Status.EXECUTING);
	}

	protected void do_expansion(AgendaItem item) {
	    Debug.noteln("Expanding using", refinement);
	    // Take the most complete env
	    MatchEnv expandEnv = filterEnv != null ? filterEnv : env;
	    // Instantiate the refinement, create subnodes, and
	    // install orderings.
	    item.expandOneLevel(refinement, expandEnv);
	    // Make any Variable bindings specified by expandEnv.
	    applyEnv(expandEnv);
	    // Add the item's children to this agenda.
	    for (Iterator i = Collect.iterator(item.getChildren());
		 i.hasNext();) {
		AgendaItem child = (AgendaItem)i.next();
		ActivityAgenda.this.addItem(child);
	    }
	    // Add issues from the refinement to the issue agenda.
	    for (Iterator i = Collect.iterator(item.refinement.getIssues());
		 i.hasNext();) {
		Issue issue = (Issue)i.next();
		ActivityAgenda.this.controller.getIssueAgenda()
		    .addItem(new IssueItem(issue));
	    }
	    // Add constraints from the refinement.
	    // /\/: At present, they are assumed to be at begin or end
	    // of self (depending on type + relation), hence apply to
	    // the activity being expanded.
	    if (item.refinement.getConstraints() != null) {
		ActivityAgenda.this.modelManager
		    .addConstraints(item, item.refinement.getConstraints());
	    }
	}

	protected void applyEnv(MatchEnv env) {
	    Map bindings = null; 	// create if needed
	    for (Iterator i = env.entrySet().iterator(); i.hasNext();) {
		Map.Entry e = (Map.Entry)i.next();
		Object var = e.getKey();
		if (var instanceof Variable) {
		    if (bindings == null)
			bindings = new HashMap();
		    bindings.put(var, e.getValue());
		}
	    }
	    if (bindings != null)
		modelManager.newBindings(bindings);
	}

    }

    /**
     * Handles an item by expanding it into subitems as specified
     * explicitly by the user.
     */
    class ManualExpandAction extends ExpandAction {

	ManualExpandAction(AgendaItem item, Refinement refinement) {
	    super(item, refinement, new MatchEnv());
	    this.shortDescription = "Expand as below";
	}

	public void handle(AgendaItem item) {
	    do_expansion(item);
	    if (item.getStatus() == Status.POSSIBLE)
		item.setStatus(Collect.isEmpty(item.getChildren())
			       ? Status.COMPLETE : Status.EXECUTING);
	}

    }


}
