/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Mar  8 01:07:15 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import javax.swing.*;

import java.awt.Container;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.*;

import java.util.*;

import ix.ip2.event.*;

import ix.iview.SimpleDomainEditor;

import ix.icore.domain.Refinement;

import ix.icore.process.PNode;

import ix.icore.*;
import ix.util.*;
import ix.util.lisp.*;

/**
 * A simple editor for entering and viewing Activities
 */
class ActivityEditor extends ItemEditor {

    ActivityEditor(AgendaViewer agendaViewer, String title) {
	super(agendaViewer, title);
    }

    protected EditPanel makeEditPanel() {
	EditPanel p = new ActivityEditPanel();
	p.setBorder(BorderFactory.createTitledBorder("Activity"));
	return p;
    }

    /**
     * Constructs the menu bar
     */
    protected JMenuBar makeMenuBar() {
        JMenuBar bar = new JMenuBar();

        JMenu fileMenu = new JMenu("File");
        bar.add(fileMenu);

        saveAsRefinementItem.setEnabled(false);
        fileMenu.add(saveAsRefinementItem);
        fileMenu.addSeparator();
        fileMenu.add(makeMenuItem("Close"));

        return bar;
    }

    /**
     * Action interpreter
     */
    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	Debug.noteln("ActivityEditor action:", command);
	if (command.equals("Save As Refinement")) {
	    Refinement r = 
		((ActivityEditPanel)editPanel).expansionPanel
		    .makePartialRefinement();
	    agendaViewer.saveExpansion(r);
	}
	else if (command.equals("Close")) {
	    frame.setVisible(false);
	}
	else
	    Debug.noteln("Nothing to do for", command);
    }

    class ActivityEditPanel extends EditPanel {

	// Buttone that can be disabled or have their text changed.
	JButton expansionButton;

	ExpansionPanel expansionPanel = new ExpansionPanel();

	ActivityEditPanel() {
	    super();
	    modifyButton = makeButton("Modify Activity");
	    expansionButton = makeButton("Specify Subactivities");
	    setUp();
	}

	protected void setUp() {

	    add(makeLeftLabel("Activity Pattern Text"));
	    add(new JScrollPane(itemText));

	    add(makeLeftLabel("Annotations"));
	    add(new JScrollPane(commentText));

	    newItemButtons.add(makeButton("Add Activity"));
	    newItemButtons.add(makeButton("Clear"));
	    newItemButtons.add(makeButton("Cancel"));

	    editItemButtons.add(modifyButton);
	    editItemButtons.add(makeButton("Copy Activity"));
	    editItemButtons.add(expansionButton);
	    editItemButtons.add(makeButton("Clear"));
	    editItemButtons.add(makeButton("Cancel"));

	    add(newItemButtons);
	}

	/**
	 * Sets up to edit an existing item.
	 */
	void showItem(AgendaItem i) {
	    ensureNoExpansionPanel();
	    Status status = i.getStatus();
	    expansionButton
		.setEnabled(status == Status.BLANK
			    || status == Status.POSSIBLE);
	    super.showItem(i);
	}

	/**
	 * Sets up to edit a new, initially empty item
	 */
	void showNewItem() {
	    ensureNoExpansionPanel();
	    super.showNewItem();
	}

	/**
	 * Action interpreter
	 */
	public void actionPerformed(ActionEvent e) {
	    String command = e.getActionCommand();
	    Debug.noteln("Activity action:", command);
	    if (command.equals("Add Activity")) {
		addItem();
	    }
	    else if (command.equals("Modify Activity")) {
		modifyItem();
	    }
	    else if (command.equals("Copy Activity")) {
		copyItem();
	    }
	    else if (command.equals("Specify Subactivities")) {
		addExpansionPanel();
	    }
	    else if (command.equals("Install Subactivities")) {
		addSubitems();
	    }
	    else if (command.equals("Modify Subactivities")) {
		throw new Error
		    ("Modify Subactivities button should be disabled");
	    }
	    else if (command.equals("Clear")) {
		showNewItem();
	    }
	    else if (command.equals("Cancel")) {
		// Equiv to close + clear.
		frame.setVisible(false);
		showNewItem();
	    }
	    else
		Debug.noteln("Nothing to do for", command);
	}

	/**
	 * Starts editing a new item based on an existing one.
	 */
	void copyItem() {
	    ensureNoExpansionPanel();
	    super.copyItem();
	}

	/**
	 * Adds subitems to an existing item.
	 */
	void addSubitems() {
	    Refinement s = expansionPanel.makePartialRefinement();
	    agendaViewer.expandItem(editingItem, s);
	}

	/*
	 * AgendaItemListener methods - in case an Item changes while
	 * we're editing it.
	 */

	public void statusChanged(AgendaItemEvent e) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("Status change while editing item", i);

	    // Since it only gets more restrictive, and always in the
	    // same way, we only have to check that we're not still in
	    // the most permissive state.
	    Status status = i.getStatus();
	    if (status != Status.BLANK && status != Status.POSSIBLE) {
		modifyButton.setEnabled(false);
		// User can look at expansion if not already doing so
		// but can't change it.
		expansionButton
		    .setEnabled(expansionButton.getText()
				.equals("Specify Subactivities"));
	    }
	    
	}

	/*
	 * Expansion / subitem panel
	 */

	/**
	 * Adds a panel for editing subitems and their ordering constraints.
	 */
	void addExpansionPanel() {
	    // Panel shouldn't already be there.
	    Debug.assert(!expansionPanelIsShown());

	    // The panel is only for editing existing items.
	    Debug.assert(editingItem != null,
			 "can't edit new item subitems");

	    expansionPanel.loadFromItem(editingItem);

	    // Allow the user to turn the item's expansion into a refinement.
	    saveAsRefinementItem.setEnabled(true);

	    if (editingItem.getChildren().isEmpty()) {
		// No existing subitems.
		expansionButton.setText("Install Subactivities");
		Status status = editingItem.getStatus();
		if (status == Status.BLANK || status == Status.POSSIBLE) 
		    expansionButton.setEnabled(true);
		else
		    expansionButton.setEnabled(false);
	    }
	    else {
		// We don't yet allow existing subitems to be modified. /\/
		expansionButton.setText("Modify Subactivities");
		expansionButton.setEnabled(false);
	    }

	    // Add the panel just above the buttons.
	    add(expansionPanel, getComponentCount() - 1);
	    frame.pack();
	}


	/**
	 * Removes the panel for editing subitems if it is being displayed.
	 */
	void ensureNoExpansionPanel() {
	    if (expansionPanelIsShown()) {
		saveAsRefinementItem.setEnabled(false);
		// The expansionButton can bring the subitem panel back
		expansionButton.setText("Specify Subactivities");
		expansionButton.setEnabled(true);
		remove(expansionPanel);
		frame.pack();
	    }
	}

	/**
	 * Determines whether the expansion panel is being displayed.
	 */
	boolean expansionPanelIsShown() {
	    return getComponent(getComponentCount() - 2) == expansionPanel;
	}


	/**
	 * A panel for editing subitems and their ordering constraints.
	 */
	class ExpansionPanel extends JPanel {

	    JTextArea subitemText = new JTextArea(5, textCols);

	    SimpleDomainEditor.TemporalConstraintPanel constraintPanel
		= new SimpleDomainEditor.TemporalConstraintPanel();

	    ExpansionPanel() {
		super();
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		// setBorder(BorderFactory.createTitledBorder("Expansion"));
		setUp();
	    }

	    void setUp() {
		add(makeLeftLabel("Subitems"));
		add(new JScrollPane(subitemText));

		add(makeLeftLabel("Constraints"));
		add(constraintPanel);
	    }

	    Refinement makePartialRefinement() {
		// Pattern
		LList pattern = 
		    Lisp.elementsFromString(itemText.getText().trim());

		// Nodes
		String expansion = subitemText.getText().trim();
		List nodes = SimpleDomainEditor.parseNodes(expansion);

		// Orderings
		String ordChoice = constraintPanel.getOrderingChoice();
		List orderings = ordChoice.equals("Sequential")
		    ? SimpleDomainEditor.sequentialOrderings(nodes)
		    : null;
		
		// Refinement
		Refinement partial = new Refinement();
		partial.setPattern(pattern);
		partial.setNodes(nodes);
		partial.setOrderings(orderings);
		partial.setComments(commentText.getText());
		return partial;
	    }

	    void loadFromItem(AgendaItem item) {
		Debug.noteln("Editing orderings of", item);
		List subnodes = item.getChildren();
		List orderings = itemOrderings(item);
		Debug.noteln("Orderings = ", orderings);

		subitemText.setText(subitemsToText(subnodes));
		constraintPanel.loadFromOrderings(subnodes, orderings);
	    }

	    String subitemsToText(List children) {
		String lineSeparator = System.getProperty("line.separator");
		StringBuffer result = new StringBuffer();
		for (Iterator n = children.iterator(); n.hasNext();) {
		    AgendaItem child = (AgendaItem)n.next();
		    result.append(Lisp.elementsToString(child.getPattern()));
		    result.append(lineSeparator);
		}
		return result.toString();
	    }

	    List itemOrderings(AgendaItem item) {
		return Lisp.NIL;

//  		// We have to turn the graph of the item's children
//  		// back into a list of orderings.  This is tedious
//  		// but straightforward.  Fortunately, we know there
//  		// are not links except to siblings.
//  		LList nodes = item.getChildren();
//  		HashMap nodeToNumber = new HashMap(nodes.size());
//  		LListCollector orderings = new LListCollector();

//  		// Give each subnode a number.
//  		int count = 1;
//  		for (Iterator n = nodes.iterator(); n.hasNext();) {
//  		    nodeToNumber.put(n.next(), new Long(count));
//  		    count++;
//  		}

//  		// Look at the direct successors of each subnode
//  		// and record a (subnode successor) ordering using
//  		// the corresponding node-numbers.
//  		for (Iterator n = nodes.iterator(); n.hasNext();) {
//  		    PNode node = (PNode)n.next();
//  		    for(Iterator p = node.postNodes.iterator(); p.hasNext();) {
//  			PNode post = (PNode)p.next();
//  			Long node_n = (Long)nodeToNumber.get(node);
//  			Long post_n = (Long)nodeToNumber.get(post);
//  			orderings.add(Lisp.list(node_n, post_n));
//  		    }
//  		}
//  		// Return list of orderings.
//  		return orderings.contents();
	    }

	}

    }

}

