/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue May 28 22:08:23 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.icore.domain.Refinement;
import ix.icore.process.ProcessModelManager;
import ix.icore.process.event.*;
import ix.icore.*;
import ix.ispace.event.*;
import ix.util.*;
import ix.util.lisp.*;

import ix.ip2.event.*;

/**
 * A collection of things to do.
 */
public abstract class Agenda implements AgendaManager,
					ProcessStatusListener,
					ContactListener {

    protected PanelController controller;

    protected ProcessModelManager modelManager; // taken from controller

    protected List items = new LinkedList();
    protected List handlers = new LinkedList();
    protected List listeners = new LinkedList();

    public Agenda(PanelController controller) {
	this.controller = controller;
	this.modelManager = controller.modelManager;
	this.modelManager.addProcessStatusListener(this);
	
	this.controller.agent.getContactManager()
	    .addContactListener(this);
    }

    public void addItem(AgendaItem item) {
	// ... subclasses may do some things first ...
	addHandlerActions(item);
	items.add(item);
	fireItemAdded(item);
    }

    public void removeItem(AgendaItem item) {
	items.remove(item);
	fireItemRemoved(item);
    }

    public void clear() {
	items.clear();
    }

    public AgendaItem makeItem(String text) {
	return makeItem(Lisp.elementsFromString(text));
    }

    public abstract AgendaItem makeItem(LList pattern);

    public void addHandlerActions(AgendaItem item) {

	// Standard options that all items have.
	// /\/: If this changes to make these come from handlers,
	// the NoteOtherReportsIssue wantsActionsFrom method must
	// change too.
	item.addAction(new HandlerAction.NoAction());
	item.addAction(new HandlerAction.Manual());
	item.addAction(new HandlerAction.NotApplicable());

	// Add a bind option if there are any unbound variables in
	// the issue pattern.
//  	{
//  	    Set vars = item.getUnboundVars();
//  	    if (!vars.isEmpty())
//  		item.addAction(controller.new BindAction());
//  	}

	// Add options from the handlers
	for (Iterator i = handlers.iterator(); i.hasNext();) {
	    ItemHandler h = (ItemHandler)i.next();
	    if (h.appliesTo(item) && item.wantsActionsFrom(h))
		h.addHandlerActions(item);
	}

    }

    public void handleItem(AgendaItem item, HandlerAction act) {
	Debug.assert(items.contains(item), "Stray item", item);
	Debug.assert(item.hasAction(act), "Stray action", act);
	act.handle(item);
    }

    public void addItemHandler(ItemHandler handler) {
	handlers.add(handler);
    }

    public void reconsiderHandler(ItemHandler handler, Object reason) {
	// /\/: See comments in PanelController.reconsiderHandler(h).
	for (Iterator i = items.iterator(); i.hasNext();) {
	    AgendaItem item = (AgendaItem)i.next();
	    if (handler.appliesTo(item)) 			// ??? /\/
		handler.reviseHandlerActions(item, reason);
	}
    }

    public void expandItem(AgendaItem item, Refinement instructions) {
	throw new Error("Attempt to expand " + item);
    }

    public boolean acceptReport(Report report) {
	for (Iterator i = items.iterator(); i.hasNext(); ) {
	    AgendaItem item = (AgendaItem)i.next();
	    if (item.wantsReport(report)) {
		item.addReport(report);
		return true;
	    }
	}
	return false;
    }

    /*
     * Methods in the ProcessStatusListener interface
     */

    // /\/: It's not clear that this is right.  Perhaps things that
    // want to know about bindings should listen to the model manager
    // directly.  Also, it might be better to have > 1 kind of MM listener,
    // because something that wants to know about bindings, e.g., might
    // not care about world-state changes.

    public void statusUpdate(ProcessStatusEvent e) {
    }

    public void newBindings(ProcessStatusEvent e, Map bindings) {
	// Pass along the event ...
	fireNewBindings(bindings);
	// Remove handler actions that are no longer valid
	checkActionValidity();
	// Recompute the status of all the handler actions
	computeActionStatus();
    }

    public void stateChange(ProcessStatusEvent e, Map delta) {
	computeActionStatus();
    }

    /*
     * Methods in the ContactListener interface
     */

    public void contactChange(ContactEvent e) {
	checkActionValidity();
    }

    /*
     * Methods to manage changes in handler actions.
     */

    protected void checkActionValidity() {
	for (Iterator i = items.iterator(); i.hasNext();) {
	    AgendaItem item = (AgendaItem)i.next();
	    checkActionValidity(item);
	}
    }

    protected void checkActionValidity(AgendaItem item) {
	boolean changes = false;
	for (Iterator i = item.getActions().iterator(); i.hasNext();) {
	    HandlerAction act = (HandlerAction)i.next();
	    if (!act.isStillValid()) {
		Debug.noteln("Removing " + act + " from " + item);
		i.remove();
		changes = true;
	    }
	}
	if (changes == true)
	    item.fireHandlerActionsChanged(); // define a diff event? /\/
    }

    protected void computeActionStatus() {
	for (Iterator i = items.iterator(); i.hasNext();) {
	    AgendaItem item = (AgendaItem)i.next();
	    computeActionStatus(item);
	}
    }

    protected void computeActionStatus(AgendaItem item) {
	// /\/: Probably should be a method of the AgendaItem,
	// but we don't want to commit too much to the
	// computeStatus way of doing things.
	Status status = item.getStatus();
	if (status == Status.COMPLETE 
	       || status == Status.EXECUTING
	       || status == Status.IMPOSSIBLE)
	    return;
	boolean changes = false;
	for (Iterator i = item.getActions().iterator(); i.hasNext();) {
	    HandlerAction act = (HandlerAction)i.next();
	    boolean oldStatus = act.isReady();
	    act.computeStatus();
	    if (act.isReady() != oldStatus) {
		changes = true;
	    }
	}
	if (changes == true)
	    item.fireHandlerActionsChanged();
    }


    /*
     * AgendaListeners and related methods.
     */

    public void addAgendaListener(AgendaListener listener) {
	listeners.add(listener);
    }

    public void fireItemAdded(AgendaItem item) {
	Debug.noteln("fireItemAdded", item);
	AgendaEvent event = new AgendaEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaListener listener = (AgendaListener)i.next();
	    listener.itemAdded(event, item);
	}
    }

    public void fireItemRemoved(AgendaItem item) {
	Debug.noteln("fireItemAdded", item);
	AgendaEvent event = new AgendaEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaListener listener = (AgendaListener)i.next();
	    listener.itemRemoved(event, item);
	}
    }

    public void fireItemHandled(AgendaItem item, HandlerAction action) {
	Debug.noteln("fireItemHandled " + item + " " +
		     action.getActionDescription());
	AgendaEvent event = new AgendaEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaListener listener = (AgendaListener)i.next();
	    listener.itemHandled(event, item, action);
	}
    }

    public void fireNewBindings(Map bindings) {
	AgendaEvent event = new AgendaEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaListener listener = (AgendaListener)i.next();
	    listener.newBindings(event, bindings);
	}
    }

}

// Issues:
// * What should be done when removing an item that has children?
//   Don't allow it?  Remove descendents as well?
