/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri May 31 00:49:01 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.icore.*;
import ix.ispace.*;
import ix.ispace.event.*;

import ix.util.*;

/**
 * Generates HandlerActions for an AgendItem.
 */
public class ForwardingHandler extends ItemHandler implements ContactListener {

    protected IXAgent agent;
    protected String action;
    protected AgentRelationship rel;
    protected boolean reportBack;

    protected ContactManager contactManager;

    public ForwardingHandler(IXAgent agent,
			     String action,
			     AgentRelationship rel,
			     boolean reportBack) {
	super(action + " to " + rel + " agents"); // action description
	this.agent = agent;
	this.action = action;
	this.rel = rel;
	this.reportBack = reportBack;

	this.contactManager = agent.getContactManager();
	this.contactManager.addContactListener(this);
    }

    public void addHandlerActions(AgendaItem item) {
	// It's assumed the controller has already determined that
	// this is an appropriate handler for the item.
	for (Iterator i = getToNames().iterator(); i.hasNext();) {
	    String to = (String)i.next();
	    item.addAction(new ForwardingAction(to));
	}
    }

    protected List getToNames() {
	List data = contactManager.getAgentData(rel);
	Debug.noteln("agent data", data);
	return data == null
	    ? Collections.EMPTY_LIST
	    : (List)Collect.map(data, Fn.accessor(AgentData.class, "getName"));
    }

    public void handle(AgendaItem item) {
	throw new Error(this + " called directly");
    }

    public void contactChange(ContactEvent e) {
	// N.B. All action removals are handled by the ForwardingAction
	// isStillValid() method, so here we handle only additions.
	if (e.isNewContact() || e.isRelationshipChange()) {
	    AgentData data = e.getNewData();
	    AgentRelationship r = data.getRelationship();
	    if (r == rel) {
		Debug.noteln("New forwarding target", data);
		((Ip2)agent).controller.reconsiderHandler(this, e);
		// Causes calls to reviseHandlerActions(i, r) for various i.
	    }
	}
    }

    public void reviseHandlerActions(AgendaItem item, Object reason) {
	// called after a relevant contact change.
	ContactEvent e = (ContactEvent)reason;
	AgentData data = e.getNewData();
	Debug.assert(data.getRelationship() == rel);

	String toName = data.getName();
	HandlerAction f = new ForwardingAction(toName);
	if (item.findAction(f.getActionDescription()) == null) {
	    item.addAction(f);
	}
    }

    public String toString() {
	return "ForwardingHandler[" + getActionDescription() + "]";
    }

    class ForwardingAction extends HandlerAction {

	String toName;

	ForwardingAction(String toName) {
	    this.toName = toName;
	    this.shortDescription = action + " to " + toName;
	}
	
	public void handle(AgendaItem item) {
	    item.getAbout().forwardTo(toName, reportBack);
	    item.setStatus(reportBack ? Status.EXECUTING : Status.COMPLETE);
        }

	public boolean isStillValid() {
	    AgentData data = contactManager.getAgentData(toName);
	    return data != null
		&& data.getRelationship() == rel;
	}

    }

}
