/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue May 21 00:27:56 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import javax.swing.*;

import java.awt.Container;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.*;

import java.util.*;

import ix.ip2.event.*;

import ix.iview.SimpleDomainEditor;
import ix.iface.util.IconImage;
import ix.iface.util.CatchingActionListener;

import ix.icore.domain.Refinement;

import ix.icore.process.PNode;

import ix.icore.*;
import ix.util.*;
import ix.util.lisp.*;

/**
 * A simple editor for entering and viewing individual items.
 */
public abstract class ItemEditor implements ActionListener {

    protected String title;
    protected JFrame frame;
    protected Container contentPane;

    protected AgendaViewer agendaViewer; // who we work for

    protected EditPanel editPanel;

    protected JMenuItem saveAsRefinementItem =
	makeMenuItem("Save As Refinement");

    ItemEditor(AgendaViewer agendaViewer, String title) {
	this.agendaViewer = agendaViewer;
	this.title = title;
	setUpFrame();
    }

    protected void setUpFrame() {
	frame = new JFrame(title);
	frame.setIconImage(IconImage.getIconImage(this));
	frame.setSize(500, 300);
	frame.setJMenuBar(makeMenuBar());

	contentPane = frame.getContentPane();

	editPanel = makeEditPanel();
	contentPane.add(editPanel);
	frame.pack();
    }

    protected abstract EditPanel makeEditPanel();

    void setVisible(boolean v) {
	frame.setVisible(v);
    }

    void showItem(AgendaItem i) {
	editPanel.showItem(i);
	setVisible(true);
    }

    void showNewItem() {
	editPanel.showNewItem();
	setVisible(true);
    }

    /**
     * Constructs the menu bar
     */
    protected abstract JMenuBar makeMenuBar();

    protected JMenuItem makeMenuItem(String text) {
	JMenuItem item = new JMenuItem(text);
	item.addActionListener(CatchingActionListener.listener(this));
	return item;
    }

    /**
     * Action interpreter.  Note that this object will be wrapped in
     * a CatchingActionListener when used as an ActionListener.
     *
     * @see ix.iface.util.CatchingActionListener
     */
    public abstract void actionPerformed(ActionEvent e);

    /**
     * Item-editing panel
     */
    protected abstract class EditPanel extends JPanel
	implements ActionListener, AgendaItemListener {

	int textCols = 50;

	JTextArea itemText = new JTextArea(3, textCols);
	JTextArea commentText = new JTextArea(6, textCols);

	// Buttons that can be disabled or have their text changed.
	JButton modifyButton;

	Box newItemButtons = Box.createHorizontalBox();
	Box editItemButtons = Box.createHorizontalBox();

	AgendaItem editingItem = null;

	EditPanel() {
	    super();
	    setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	    // setUp(); // done in leaf subclass constructors
	}

	protected abstract void setUp();

	protected Box makeLeftLabel(String text) {
	    Box b = Box.createHorizontalBox();
	    b.add(new JLabel(text));
	    b.add(Box.createHorizontalGlue());
	    return b;
	}

	protected JButton makeButton(String text) {
	    JButton button = new JButton(text);
	    button.addActionListener(CatchingActionListener.listener(this));
	    return button;
	}

	/**
	 * Sets up to edit an existing item.
	 */
	void showItem(AgendaItem i) {

	    editingItem = i;
	    i.addItemListener(this);

	    itemText.setText(Lisp.elementsToString(i.getPattern()));
	    itemText.setEnabled(false);

	    commentText.setText(i.comments);
	    commentText.setEnabled(true);

	    ensureButtons(editItemButtons);

	    Status status = i.getStatus();
	    if (status == Status.BLANK || status == Status.POSSIBLE) {
		modifyButton.setEnabled(true);
	    }
	    else {
		// /\/: Instead allow modification of e.g. comments?
		modifyButton.setEnabled(false);
		commentText.setEnabled(false);
	    }

	    frame.validate();

	}

	/**
	 * Sets up to edit a new, initially empty item
	 */
	void showNewItem() {

	    editingItem = null;

	    itemText.setText("");
	    itemText.setEnabled(true);

	    commentText.setText("");
	    commentText.setEnabled(true);

	    ensureButtons(newItemButtons);
	    frame.validate();

	}

	void ensureButtons(Box buttons) {
	    Component lastComponent = getComponent(getComponentCount() - 1);
	    Debug.assert(lastComponent == newItemButtons
			 || lastComponent == editItemButtons);
	    if (lastComponent != buttons) {
		remove(lastComponent);
		add(buttons);
		repaint();
	    }
	}


	/**
	 * Action interpreter.  Note that this object will be wrapped in
	 * a CatchingActionListener when used as an ActionListener.
	 *
	 * @see ix.iface.util.CatchingActionListener
	 */
	public abstract void actionPerformed(ActionEvent e);


	/**
	 * Tells the item viewer to add a new item
	 */
	void addItem() {
	    Debug.assert(editingItem == null);
	    String text = itemText.getText().trim();
	    String comments = commentText.getText();
	    if (text.length() == 0) {
		JOptionPane.showMessageDialog(frame,
	            "Empty item text",
		    "Invalid item",
		    JOptionPane.ERROR_MESSAGE);
		return;
	    }
	    try {
		LList pattern = Lisp.elementsFromString(text);
		if (pattern.length() < 1)
		    throw new RuntimeException("Empty item pattern");
		AgendaItem i = agendaViewer.makeItem(pattern);
		i.setComments(comments);
		agendaViewer.addItem(i);
		showItem(i);	// now editing that item
	    }
	    catch (Exception e) {
		Debug.noteException(e);
		JOptionPane.showMessageDialog(this,
	            new Object[] {"Cannot define item.", e.getMessage()},
		    "Error while defining item",
		    JOptionPane.ERROR_MESSAGE);
		return;
	    }
	}

	/**
	 * Modifies a existing item.
	 */
	void modifyItem() {
	    Debug.assert(editingItem != null);
	    editingItem.setComments(commentText.getText());
	}

	/**
	 * Starts editing a new item based on an existing one.
	 */
	void copyItem() {
	    Debug.assert(editingItem != null);
	    if (!editingItem.getPatternVars().isEmpty()) {
		// Go back from Variables to ?-vars.
		LList pat = (LList)Variable
		    .removeVars(editingItem.getPattern());
		itemText.setText(Lisp.elementsToString(pat));
	    }
	    editingItem = null;
	    itemText.setEnabled(true);
	    commentText.setEnabled(true);
	    ensureButtons(newItemButtons); // can't edit subactions
	    frame.validate();
	}

	/*
	 * AgendaItemListener methods - in case an Item changes while
	 * we're editing it.
	 */

	public void statusChanged(AgendaItemEvent e) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("Status change while editing item", i);

	    // Since it only gets more restrictive, and always in the
	    // same way, we only have to check that we're not still in
	    // the most permissive state.
	    Status status = i.getStatus();
	    if (status != Status.BLANK && status != Status.POSSIBLE) {
		modifyButton.setEnabled(false);
	    }
	    
	}

	public void priorityChanged(AgendaItemEvent e) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("Priority change while editing item", i);
	}

	public void handlerActionsChanged(AgendaItemEvent e) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("Handler actions changed while editing item", i);
	}

	public void newHandlerAction(AgendaItemEvent e, HandlerAction act) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("New handler action while editing item", i);
	}

	public void newReport(AgendaItemEvent e, Report report) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    Debug.noteln("New report while editing item", i);
	    commentText.setText(i.comments);
	    // Try to get it to scoll to the end
	    commentText.setCaretPosition(i.comments.length());
	    frame.validate();
	}

	public void agendaItemEdited(AgendaItemEvent e) {
	    AgendaItem i = (AgendaItem)e.getSource();
	    if (i != editingItem)
		return;
	    // Probably we did the editing ...
	    Debug.noteln("Item edited while editing", i);
	}

    }

}

