/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May 30 17:56:22 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.ip2.event.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.icore.domain.event.*;
import ix.icore.process.ProcessModelManager;

import ix.util.*;

/**
 * Controller for process and other, similar panels.
 */
public class PanelController implements DomainListener {

    protected IXAgent agent;
    Domain domain;		// /\/: ref in ActivityAgenda

    protected List listeners = new LinkedList();

    // /\/: Should this be here or in the IXAgent?
    // N.B. value needed when agendas created.
    ProcessModelManager modelManager = new Ip2ModelManager();

    // /\/: Can't be init here bec agenda refers to agent
    // to get its contact manager
    protected ActivityAgenda activities;
    protected Agenda issues;

    public PanelController(IXAgent agent) {
	this.agent = agent;
	this.activities = new ActivityAgenda(this);
	this.issues = new IssueAgenda(this);
    }

    public void reset() {
	modelManager.reset();
	activities.clear();
	issues.clear();
    }

    public void setDomain(Domain domain) {
	this.domain = domain;
	this.domain.addDomainListener(this);
    }

    public Agenda getActivityAgenda() {
	return activities;
    }

    public Agenda getIssueAgenda() {
	return issues;
    }

    /*
     * Incoming events.
     */

    public void addActivity(Activity activity) {
	if (activity instanceof WaitForReportsActivity)
	    // /\/: irritating special case
	    activities.addItem(new WaitForReportsItem(activity));
	else
	    activities.addItem(new ActivityItem(activity));
    }

    public void addIssue(Issue issue) {
	issues.addItem(new IssueItem(issue));
    }

    public void newReport(Report report) {
	// N.B. Must try activities first because of "note other reports"
	// issue.
	if (!(activities.acceptReport(report)
	      || issues.acceptReport(report)))
	    throw new Error("Failed to accept " + report);
    }

    /*
     * Handler-related methods.
     */

    public void addActivityHandler(ItemHandler handler) {
	activities.addItemHandler(handler);
    }

    public void addIssueHandler(ItemHandler handler) {
	issues.addItemHandler(handler);
    }

    public void addItemHandler(ItemHandler handler) {
	activities.addItemHandler(handler);
	issues.addItemHandler(handler);
    }

    public void reconsiderHandler(ItemHandler handler, Object reason) {
	// Called by a handler when it things it may need to
	// revise some AgendaItems' lists of handler actions.
	// /\/: Somewhat awkward that it has to go via the controller,
	// but a handler may be attached to more than one agenda
	// and doesn't have a direct way to reach any of them.

	Debug.noteln("Controller asked to reconsider", handler);
	Debug.noteln("For reason", reason);

	if (issues.handlers.contains(handler))
	    issues.reconsiderHandler(handler, reason);
	if (activities.handlers.contains(handler))
	    activities.reconsiderHandler(handler, reason);
    }

    /*
     * Adding listeners to the agendas.
     */

    public void addActivityListener(AgendaListener listener) {
	activities.addAgendaListener(listener);
	listener.setAgendaManager(activities);
    }

    public void addIssueListener(AgendaListener listener) {
	issues.addAgendaListener(listener);
	listener.setAgendaManager(issues);
    }

    /*
     * Methods required by the DomainListener interface.
     */

    public void refinementAdded(RefinementEvent event) {
	Refinement r = event.getRefinement();
	Debug.noteln("Panel Controller noting new ", r);
	activities.refinementAdded(r);
    }

    /*
     * Methods for dealing with ControllerListeners.
     */

    public void addControllerListener(ControllerListener listener) {
	listeners.add(listener);
    }

    /*
     * HandlerAction classes
     */

//  public class BindAction extends HandlerAction {

//  	BindingViewer binder;

//  	public BindAction() {
//  	    this.shortDescription = "Bind variables";
//  	}

//  	public void handle(AgendaItem item) {
//  	    Set unboundVars = item.getUnboundVars();
//  	    if (unboundVars.isEmpty())
//  		throw new RuntimeException("No variables to bind.");

//  	    binder = new BindingViewer
//  		(agent,
//  		 unboundVars,
//  		 new Runnable() {
//  		     public void run() {
//  			 newBindings();
//  		     }
//  		 });
//  	}

//  	void newBindings() {
//  	    // Here's where we actually assign the values.
//  	    Map bindings = binder.newBindings;
//  	    for (Iterator i = bindings.entrySet().iterator();
//  		 i.hasNext();) {
//  		Map.Entry e = (Map.Entry)i.next();
//  		Variable var = (Variable)e.getKey();
//  		Object val = e.getValue();
//  		var.setValue(val);
//  	    }
//  	    PanelController.this.activities.fireNewBindings(bindings);
//  	    PanelController.this.issues.fireNewBindings(bindings);
//  	    // /\/: Presumably will have to tell state viewers too.
//  	}

//  }

}
