/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May 30 04:01:14 2002 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.itest;

import java.awt.Container;
import java.awt.event.*;

import javax.swing.*;

import java.util.*;

import java.io.File;		// for test-menu parameter /\/

import ix.iface.util.LogoPanel;

import ix.ip2.TestItem;

import ix.icore.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.xml.*;		// for test-menu parameter /\/


/**
 * The generic I-TEST class and application main program. <p>
 *
 * Its IPC name can be given by the "-ipc-name" command-line argument.
 * For example, to pretend to be I-DEEL:
 * <pre>
 *    java ix.itest.Itest -ipc -ipc-name=IDEEL
 * </pre>
 *
 */
public class Itest extends IXAgent {

    protected ItestFrame frame;

    protected SendPanel sendPanel;
    protected ReceivePanel receivePanel;

    public Object getAgentIPCName() { return ipcName; }

    public Itest() {
	this("I-Test");
    }

    public Itest(String frameTitle) {
	displayName = frameTitle;
    }

    /**
     * Main program.
     */
    public static void main(String[] argv) {

	Util.printGreeting("I-Test");

	new Itest().mainStartup(argv);

    }

    /**
     * Command-line argument processing used by all versions of I-TEST.
     *
     * @see ix.icore.IXAgent#processCommandLineArguments()
     */
    protected void processCommandLineArguments() {
	super.processCommandLineArguments();

//  	displayName = Parameters.haveParameter("display-name")
//  	    ? Parameters.getParameter("display-name")
//  	    : getAgentSymbolName() + " I-Test";

	if (Parameters.haveParameter("test-menu"))
	    // Avoids warning about unused parameter. /\/
	    Debug.noteln("Test menu file",
			 Parameters.getParameter("test-menu"));
    }

    /**
     * Completes I-TEST setup and initialization.
     */
    public void startup() {
	super.startup();

	frame = new ItestFrame(displayName);
	sendPanel = new SendPanel(this);
	receivePanel = new ReceivePanel(this);

	setupFrame();
	addTests();
	frame.pack();
	frame.setVisible(true);
    }

    /**
     * Gives the main I-TEST frame its contents
     */
    public void setupFrame() {
	Container contentPane = frame.getContentPane();
	contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
	contentPane.add(sendPanel);
	contentPane.add(receivePanel);
	contentPane.add(makeLogoPanel());
    }

    /**
     * Creates a standard logo panel for I-TEST.
     * Overridden in subclasses that have, say, a picture to use instead.
     */
    public JPanel makeLogoPanel() {
	return new LogoPanel("I-Test", "I-X Test Agent",
			     Util.resourceImageIcon("ip2-itest-logo.gif"));
    }


    /**
     * Handles external input in the form of an IPC.InputMessage
     * that contains an Issue or Report object.
     */
    public void handleInput(IPC.InputMessage message) {
	receivePanel.handleInput(message);
    }


    /**
     * Adds any tests common to all versions of I-TEST.
     */
    protected void addTests() {
	if (Parameters.haveParameter("test-menu"))
	    addTests(Parameters.getParameter("test-menu"));
    }

    protected void addTests(String filename) {
	// /\/: Essentially duplicates code in Ip2 class.
	try {
	    List items = (List)XML.objectFromFile(new File(filename));
	    if (items != null) {
		for (Iterator i = items.iterator(); i.hasNext();) {
		    TestItem item = (TestItem)i.next();
		    addTest(item);
		}
	    }
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    Util.displayMessage(null,
	        Strings.foldLongLine
		  ("Problem parsing " + Strings.quote(filename) + ": " + e));
	}
    }

    /*
     * Test utilities
     */

    /**
     * Adds a test that sends a single issue to a designated agent.
     *
     * @deprecated  As of May 2002, use {@link #addTest(TestItem)}
     * instead.
     */
    protected void addIssueTest(String toName, String issueText) {
	addIssueTest(toName, Priority.NORMAL, issueText);
    }

    /**
     * Adds a test that sends a single issue to a designated agent.
     *
     * @deprecated  As of May 2002, use {@link #addTest(TestItem)}
     * instead.
     */
    protected void addIssueTest(final String toName,
				Priority priority,
				String issueText) {
	String buttonText = "Send to " + toName + " " + Util.quote(issueText);

	final Issue issue = new Issue(issueText);
	issue.setPriority(priority);
	issue.setSenderId(Name.valueOf((String)getAgentIPCName()));

	frame.addTest(buttonText, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
		IPC.exceptionlessSend(toName, issue);
	    }
	});
    }

    /**
     * Adds a test menu item based on a TestItem.
     */
    public void addTest(TestItem item) {
	final Sendable contents = item.getContents();
	final String toName = item.getToName();
	String menuText = Strings.replace("$to", toName, item.getMenuText());

	contents.setSenderId(Name.valueOf((String)getAgentIPCName()));

	frame.addTest(menuText, new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		// N.B. must make a copy, because it will change
		// its status, etc.
		sendPanel.sendCopy(toName, contents);
	    }
	});

    }

    /**
     * Adds a test that sends a series of reports to a designated agent.
     */
    protected void addReportSequenceTest(String buttonText,
					 String toName,
					 String ref,
					 List messageTexts) {
	frame.addTest(buttonText,
		      new TestSequenceListener(toName, ref, messageTexts));
    }

    /**
     * Adds a test that sends a series of reports to a designated agent.
     */
    protected void addReportSequenceTest(String buttonText,
					 String toName,
					 String ref,
					 String[] messageTexts) {
	addReportSequenceTest
	    (buttonText, toName, ref,
	     Seq.toLList(Seq.elements((Object[])messageTexts)));
    }

    /**
     * ActionListener for AbstractButtons that send a sequence of messages.
     */
    public class TestSequenceListener implements ActionListener {
	SendingThread sender;
	String toName;
	Name ref;
	List messageTexts;
	List reports;

	public TestSequenceListener(String toName,
				    String ref,
				    List messageTexts) {
	    this.toName = toName;
	    this.ref = Name.valueOf(ref);
	    this.messageTexts = messageTexts;
	    this.reports = makeReports(messageTexts);
	}

	public void actionPerformed(ActionEvent event) {
	    AbstractButton button = (AbstractButton)event.getSource();
	    String command = event.getActionCommand();
	    if (command.startsWith("Send")) {
		sender = new SendingThread(button, toName, reports);
		sender.start();
	    }
	    else if (command.startsWith("Stop")) {
		sender.exit = true;
		Debug.noteln("Stopped messages to", toName);
	    }
	    else
		Debug.noteln("Strange test button action", command);
	}

	public List makeReports(List messageTexts) {
	    List result = new LinkedList();
	    for (Iterator i = messageTexts.iterator(); i.hasNext();) {
		String text = (String)i.next();
		Report report = new Report(text);
		report.setSenderId(Name.valueOf((String)getAgentIPCName()));
		report.setRef(ref);
		result.add(report);
	    }
	    return result;
	}

    }


    /**
     * Stoppable thread for sending a series of messages.
     * Stop by doing <code>thread.exit = true</code>.
     */
    public static class SendingThread extends CatchingThread {

	AbstractButton button;
	Object destination;
	List messages;

	// These delays should be at a higher level so that its
	// easy to specify them in the code that creates the test.  /\/
	public int defaultDelay = 2000;	      // milliseconds between messages
	public int initialDelay = 10*1000;    // extra delay before 1st message

	volatile boolean exit = false; // gets the thread to exit.

	public SendingThread(AbstractButton button,
			     Object destination,
			     List messages) {
	    this.button = button;
	    this.destination = destination;
	    this.messages = messages;
	    // We assume the thread will be started pretty much
	    // right away.
	    changeButtonVerb(button, "Send", "Stop sending");
	}

	public void innerRun() {
	    Debug.noteln("Sending to " + destination, messages);
	    List sent = new LListCollector();

	    // Initial delay (need for screen cams)
	    try { Thread.sleep(initialDelay); }
	    catch (InterruptedException e) {}

	messageLoop:
	    for (Iterator i = messages.iterator();
		 !exit && i.hasNext();) {
		Object message = i.next();
		Debug.noteln("Sending", message);
		try { IPC.sendObject(destination, message); }
		catch (IPC.IPCException e) {
		    Debug.noteln("Stopped sending because " + e.getMessage());
		    break messageLoop;
		}
		sent.add(message);
		try { Thread.sleep(defaultDelay); }
		catch (InterruptedException e) {}
	    }
	    Debug.noteln("Sent to " + destination, sent);
	    SwingUtilities.invokeLater(new Runnable() {
		public void run() {
		    changeButtonVerb(button, "Stop sending", "Send");
		}
	    });
	}

    }

    /**
     * Changes the an initial segment of an abstract button's text.
     */
    public static void changeButtonVerb(AbstractButton button,
				 String from, String to) {
	String text = button.getText();
	Debug.assert(text.startsWith(""),
		     "expected " + Util.quote(from) +
		     " to begin " + Util.quote(text));
	String newText = to + text.substring(from.length());
	button.setText(newText);
    }

}
