/****************************************************************************
 * An editor/viewer panel for domain structures
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iview;

import java.lang.Math;
import java.util.Enumeration;
import java.util.List;
import javax.swing.*;       
import java.awt.*;
import java.awt.event.*;
import ix.*;
import ix.util.Debug;
import ix.util.Util;
import ix.util.lisp.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.iface.ui.*;
import ix.iview.*;
import ix.iview.domain.*;

/****************************************************************************
 * An editor/viewer panel for named domain constructs.
 *
 * The editor will not let you update a current construct yet, because we have
 * not yet worked out how to manage the impact this would have on other 
 * modules that use the domain. 
 * The editor will let you add new constructs (if you have changed an old one,
 * change the name to save it).<p>
 *
 * The construct name, pattern, and comments are currently simply strings. They
 * can all be edited by typing into their fields. The pattern should not stay
 * as a string - it should change into something a bit more constrained.
 * This will be changed when the construct models have been sorted out.<p>
 * 
 * Nodes and constraints can be added, deleted, and cleared, either
 * using the buttons or using keys (enter or insert for add, delete for
 * delete).<p>
 *
 * Nodes are simple LLists for the moment. Adding a node
 * expects the user to type in a pattern (currently just a string) and it will 
 * generate a number for the node. Deleting nodes will also
 * delete any constraints in which the deleted node participates.<p>
 *
 * The only constraints currently known are ordering constraints. Adding
 * constraints will bring up an editor that lets the user select activities
 * and time-points. For the moment, these "proper" constraints are translated
 * into simple versions because other modules cannot handle anything more
 * complex. A simple constraint "(<before> <after>)" stands for
 * "(<before> End Before Begin <after>). The constraint editor and the
 * DomainEditor are ready to provide and manage "proper" constraints, but we
 * will not provide them until other modules can deal with them.<p>
 *
 * Example code for using the DomainEditorPanel:
 *<PRE><code>
 *    ...
 *    editorPanel = new DomainEditorPanel(this);
 *    mainJPanel.add(editorPanel); 
 *    pack();
 *    editorPanel.setUIDomain(uiDomain);
 *    ...
 *</code></PRE>
 *****************************************************************************
 */
public abstract class AConstructFramePanel extends JPanel 
  implements ConstructFraming, ActionListener
{

  //---------------------Fields-------------------------------------------
  /** The panels parent - a DomainEditor */
  public DomainEditor parent;
  protected JScrollPane scrollPane = new JScrollPane();
  protected JPanel mainPanel;
  protected IFormPanel formPanel;

  private String alwaysSave = "";
  private boolean switching = false;

  //---------------------Constructors----------------------------------------

  /**
   * Creates an editor panel with the given parent.
   * Creates all panel components, the mouse listener, and the relation editor.
   *
   * @param parent the DomainEditor frame that the panel belons to. Used 
   *   mainly to keep toggle buttons and menus in synch
   */    
  public AConstructFramePanel(){
    super();
  }
  public AConstructFramePanel(DomainEditor theParent){
    //make all things (some are made during initialisation)
    super();
    parent = theParent;
    setupEditor();
    //newConstruct();
  }

  //---------------------Window building-------------------------------------
  /**
   * Sets the panel's layout and dimensions and the main panel's layout.
   */
  protected void setupEditor() {
    setupOuterPanel();
    mainPanel = setupMainPanel();
    scrollPane.getViewport().add(mainPanel); //mainPanel
  }

  protected void setupOuterPanel() {
    this.setLayout(new BorderLayout());
    //scrollPane.setBounds(0,0,405,505);
    this.add(scrollPane,BorderLayout.CENTER);
  }
  abstract protected JPanel setupMainPanel();


  public void registerListKeys(JComponent comp, String commandEnd) {
    comp.registerKeyboardAction(this,"add " + commandEnd,
				KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,0),
				comp.WHEN_FOCUSED);
    comp.registerKeyboardAction(this,"add " + commandEnd,
				KeyStroke.getKeyStroke(KeyEvent.VK_INSERT,0),
				comp.WHEN_FOCUSED);
    comp.registerKeyboardAction(this,"delete " + commandEnd,
				KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,0),
				comp.WHEN_FOCUSED);
  }

  //---------------------Public services--------------------------------------

  /**
   * Sets the given construct to be viewed/edited. 
   * Makes a UI object unless there is already one, displays its details in the
   * relevant fields of the editor, and remembers the UI object as current.
   *
   * @param object the object to be viewed/edited
   */    
  public void setConstruct(IXObject object) {
    ((ConstructEditing)mainPanel).setConstruct(object);
   }
  /**
   * Sets the given UI construct to be viewed/edited. 
   * Displays the ojbect's details in the relevant fields of the editor, 
   * and remembers the object as current.
   *
   * @param object the UIObject to be viewed/edited
   */    
  public void setUIConstruct(UIObject object) {
    ConstructEditing editPanel = (ConstructEditing)mainPanel;
    //if (object.equals(editPanel.getUIConstruct())) 
    //  return;
    //set the object in the actual panel
    UIObject current = editPanel.getUIConstruct();
    if ((current != null) && current.equals(object)) 
      editPanel.loadFromObject(); //in case object changed
    else editPanel.setUIConstruct(object);
   }


  public Domain getDomain() {
    return getUIDomain().getDomain();
  }
  public UIDomain getUIDomain() {
    return parent.getUIDomain();
  }
  public void setUIDomain(UIDomain uid) {
    if (uid.equals(getUIDomain())) return;
    else parent.setUIDomain(uid);
  }

  public void clear() {
    newConstruct();
  }

  public boolean beforeClosing() {
    return beforeSwitching();
  }
  public boolean beforeOpening() {
    if (ConstructEditing.class.isInstance(mainPanel))
      setUIConstruct(((ConstructEditing)mainPanel).getUIConstruct());
    return true;
  }

  /**
   * Notes the current panel's contents in the current construct before 
   * switching. 
   * @return true if the switch should still take place, false to cancel the
   * switch.
   */
  public boolean beforeSwitching() {
    //if (switching) return true;

    switching = true;
    //Debug.noteln("ACFP: Preparing for switch");
    if (alwaysSave.equals("always")) {
      noteConstruct();
      switching = false;
      return true;
    }
    if (alwaysSave.equals("never")) {
      switching = false;
      return true;
    }
    
    UIObject uio = ((ConstructEditing)mainPanel).getUIConstruct();
    if (uio == null) uio = makeNewConstruct();
    ((ConstructEditing)mainPanel).saveToObject(uio);
    ((ConstructEditing)mainPanel).setUIConstruct(uio);

    //if there is no object or it has not changed since the last save,
    // just switch
    //Debug.noteln("ACFP: checking uio for change - uio:", uio.getName());
    if (!uio.hasChanged()) {
      switching = false;
      return true;
    }
    //Debug.noteln(" uio has changed");
    
    String[] message = 
      {"Modify draft?",
       "(You have made changes in the panel)",
       "Do you want to",
       "- Modify: put the changes into the draft, then continue",
       "- Discard: continue without noting changes",
       "- Cancel: Do nothing",
       "- Always Modify: Always put changes into the draft",
       "- Always Discard: Always discard changes",
      };
    Object[] options = {"Modify", "Discard" , "Cancel" , 
			"Always Modify", "Always Discard"};
    int y = JOptionPane.showOptionDialog(null, message, "Confirm", 
					 JOptionPane.DEFAULT_OPTION, 
					 JOptionPane.QUESTION_MESSAGE,
					 null, options, options[0]);
    if (y == 0) { //save
      noteDraftConstruct(uio);
      switching = false;
      return true;
    }
    else if (y == 1) { //discard
      //reset the construct to its last saved state, then switch
      uio.loadFromDomain();
      //Debug.noteln("ACFP: discarded change - name is now", uio.getName());
      ((ConstructEditing)mainPanel).setUIConstructOnly(uio);
      ((ConstructEditing)mainPanel).loadFromObject();
      switching = false;
      return true;
    }
    else if (y == 2) {
      //Debug.noteln("ACFP: cancel switch");
      switching = false;
      return false; //cancel
    }
    else if (y == 3) { //always save
      alwaysSave = "always";
      noteDraftConstruct(uio);
      switching = false;
      return true;
    }
    else if (y == 4) { //always discard
      alwaysSave = "never";
      //reset the construct to its last saved state, then switch
      uio.loadFromDomain();
      ((ConstructEditing)mainPanel).setUIConstructOnly(uio);
      ((ConstructEditing)mainPanel).loadFromObject();
      switching = false;
      return true;
    }
    else {
      switching = false;
      return false;    
    }
  }

  /**
   * Note the UIObject in the (right) draft domain
   */
  protected void noteDraftConstruct(UIObject uio) {
    uio.noteChange();
    //make sure the construct is noted into the right domain
    if (getUIDomain().equals(uio.getUIDomain())) 
      uio.moveToUIDomain(getUIDomain());
    getUIDomain().updateConstruct(uio);
    //Debug.noteln("ACFP: Noted construct");
    //Debug.noteln(" ",uio.print());
  }

  //---------------------Abstract services------------------------------------

  abstract public void newConstruct();
  abstract public UIObject makeNewConstruct();
  
  abstract public boolean frameActionPerformed(ActionEvent event);
  /**
   * Wakes up the ActionListener with a user action.
   * This is called when a KeyStroke happens in which the ActionListener
   * registered its interest.
   */
  abstract public void actionPerformed(ActionEvent ae);
  /**
   * Note the contents of the panel into the current object.
   */
  abstract protected UIObject noteConstruct();

}

/*Issues***************************************************************
 *
/*Todos***************************************************************
 *
 * - 
 *  Wait for 
 *
 *****************************************************************************
 */
