/****************************************************************************
 * An editor for specifying conditions and effects in refinements.
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */

package ix.iview;

import javax.swing.*;       
import java.util.*;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import ix.*;
import ix.util.*;
import ix.util.Debug;
import ix.util.lisp.*;
import ix.icore.domain.*;
import ix.iview.domain.*;
import ix.iface.ui.*;
import ix.iface.domain.*;
import ix.ip2.Ip2ModelManager;

/****************************************************************************
 * An editor for specifying conditions and effects in refinements.
 *
 * Conditions and effects are attached to self (the refinement) or to 
 * a child. <p>
 *
 * The editor is implemented as a frame using JList objects, a JCheckBox
 * and JButtons.  <p>
 *
 * The editor can be created with no arguments or with a given refinement
 * which will be set as the current refinement whose conditions and effects
 * are to be specified. Subsequently, the refinement in the editor can be set
 * using the setRefinement method.
 *****************************************************************************
 */
public class ConditionEffectEditor extends ConstraintEditor
{
  //relation items
  

  protected JList actList;
  protected JRadioButton condBut;
  protected JRadioButton effBut;
  private ButtonGroup group;
  protected JTextField statement;
  LTF_Parser constraintParser = new LTF_Parser();
  
  /**
   * Creates an editor for specifying time point relations.
   * The editor has components for selecting two time points of two
   * sub-activities, and two buttons (Ok, Cancel)
   * "- enter pattern-value statements,",
   * "- select whether you are specifying a condition or an effect,",
   * "- select the node to which the condition/effect applies."});
   */
  public ConditionEffectEditor(){
    super("Condition/Effect Editor",
	  "Please specify a condition or an effect");
    }
    
  /**
   * Creates an editor for specifying time point relations within the given
   * refinement.
   * As OrderingEditor() but with a given refinement.
   * @param refinement the refinement whose sub-activities are to be related
   */
  public ConditionEffectEditor(UIRefinement refinement) {
    this();
    setObject(refinement);
  }
  
  
  /**
   * Makes all items on the relations panel.
   * These are two identical activity lists and two identical time point lists
   * and a "Before" label between the two
   */
  protected Component makeRels() {
    //make a panel that can take all the bits
    JPanel panel = new JPanel(new GridBagLayout());
    
    //make the activity list panel
    actList = new JList();    
    ListEditorPanel actPanel = makeActPanel(actList);
    
    //make the statement text
    JLabel label = new JLabel("Statement (pattern and value)");
    statement = new JTextField();

    //make the condition/effects buttons
    condBut = new JRadioButton("Condition", true);
    effBut = new JRadioButton("Effect", false);
    condBut.setActionCommand("Condition");
    effBut.setActionCommand("Effect");
    Object[] typeButs = {condBut, effBut};
    IXButPanel typeButPanel = new IXButPanel(this, BoxLayout.Y_AXIS, typeButs);
    typeButPanel.setButtonGroup(true);
    group = typeButPanel.group;
    
    //add it all to the panel
    GridBagConstraints c;
    c = new GridBagConstraints(0, 0, //x,y
			       2, 1, //width height
			       0.0,0.0,  //weight x,y
			       GridBagConstraints.WEST, //anchor
			       GridBagConstraints.NONE,  //fill
			       new Insets(0,2,0,2),0,0); //insets and padding
    panel.add(label, c);
    c = new GridBagConstraints(0, 1, //x,y
			       2, 1, //width height
			       1.0,0.0,  //weight x,y
			       GridBagConstraints.WEST, //anchor
			       GridBagConstraints.HORIZONTAL,  //fill
			       new Insets(0,2,0,2),0,0); //insets and padding
    panel.add(statement, c);
    c = new GridBagConstraints(0, 2, //x,y
			       1, 2, //width height
			       1.0,1.0,  //weight x,y
			       GridBagConstraints.WEST, //anchor
			       GridBagConstraints.BOTH,  //fill
			       new Insets(0,2,0,2),0,0); //insets and padding
    panel.add(actPanel, c);
    c = new GridBagConstraints(1, 2, //x,y
			       1, 2, //width height
			       0.0,0.0,  //weight x,y
			       GridBagConstraints.EAST, //anchor
			       GridBagConstraints.NONE,  //fill
			       new Insets(0,2,0,2),0,0); //insets and padding
    panel.add(typeButPanel, c);

    return panel;
  }

  /*
   * Make a sub-panel with the given activity list to be filled later
   * @param list the JList that will hold the activities
   * Uses a NodeNumberRenderer to display the list items
   * @return the new panel
   */
  private ListEditorPanel makeActPanel(JList list){
    ListEditorPanel panelAct = new ListEditorPanel(this,"Activity",list);
    list.setAutoscrolls(true);
    list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    //list.setBackground(Color.white);
    list.setCellRenderer(new NodeNumberRenderer(false));
    return panelAct;
  }


  /**
   * Sets the refinement whose components (children) are to be realted.
   * Fills the activity lists of the panel with the children of the refinement.
   * @param refinement the refinement whose children are to be related 
   *
   * @throws NullPointerException if the refinement has no children.
   */
  protected void initFromObject() throws NullPointerException {
    UIRefinement refinement = (UIRefinement) currentObject;
    if (refinement == null) return;
    List children = refinement.getNodes();
    children.add(0, refinement);
    // Debug.noteln("Children are", children);
    this.makeActList(this.actList, children.toArray());
    children.remove(0);
  }

  protected void setFromObject(Object original) {
  }

  private void makeActList(JList list, Object[] children){
    list.removeAll();
    list.setListData(children);
  }

  /**
   * Reads the selected relation and let interested parties know.
   * Checks that selections have been made.
   */
   protected Object collectConstraint() {
    
    Object node = actList.getSelectedValue();
    ButtonModel bm = group.getSelection();
    Debug.noteln("Selected button is",bm);
    
    String type = bm.getActionCommand(); //Condition or Effect
    Debug.noteln("Type is", type);
    Symbol condEff;
    if (type.equals("Effect"))
      condEff = Refinement.S_EFFECT;
    else condEff = Refinement.S_CONDITION;
    String text = statement.getText().trim();
    Debug.noteln("Got statement", text);
    List parameters = readCondition(text);
    if (isEditing()) {
      Constraint constraint = (Constraint)original;
      constraint.setRelation(condEff);
      constraint.setParameters(parameters);
      return null;
    }
    else return new Constraint(Refinement.S_WORLD_STATE, condEff, parameters);
  }

  
    protected String ensureParenthesized(String text) {
	if (text.startsWith("(") && text.endsWith(")"))
	    return text;
	else
	    return "(" + text + ")";
    }
  protected List readCondition(String text) {
    String[] parts = Strings.breakAtFirst("=", text);
    String pat = ensureParenthesized(parts[0].trim());
    String val = parts[1].trim();
    if (val.equals("")) val = "true";

	String parameters = "(" + pat + " = " + val + ")";
	List spec = (LList)Lisp.readFromString(parameters);
	return spec;
    }

  public void mouseClicked(MouseEvent e){
    if (e.getSource() == actList) {
      int selected = actList.getSelectedIndex();
      if (selected != 0) {
	String message = "Sofar, conditions and effects can only be" +
	  " attached to refinements (i.e. self)";
	JOptionPane.showMessageDialog(this,message);
	actList.setSelectedIndex(0);
      }
    }
    else super.mouseClicked(e);
  }
  
}

//This would probably be neater with JComboBox objects but that didn't work.
// see RelationEditor for attempts.

