/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Feb  8 19:21:05 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.lang.reflect.*;
import java.util.*;

/**
 * Copies Objects, including to instances of a different class.
 */
public class ObjectCopier {

    protected Map walkerCache = new HashMap();

    public ObjectCopier() { }

    /**
     * Returns a top-level copy of the object.
     */
    public Object copy(Object obj)
	   throws InstantiationException,
	          IllegalAccessException,
		  InvocationTargetException {
	return copy(obj, obj.getClass());
    }

    /**
     * Returns a top-level copy as an instance of the specified result
     * class.
     */
    public Object copy(Object obj, Class resultClass)
	   throws InstantiationException,
		  IllegalAccessException,
		  InvocationTargetException {
	Object result = resultClass.newInstance();
	FieldWalker ow = ensureFieldWalker(obj.getClass());
	FieldWalker rw = ensureFieldWalker(resultClass);
	for (Iterator fi = ow.getFieldInfo().iterator(); fi.hasNext();) {
	    FieldWalker.FieldInfo ofield = (FieldWalker.FieldInfo)fi.next();
	    FieldWalker.FieldInfo rfield = rw.getFieldInfo(ofield.name);
	    if (rfield != null)
		rfield.setValue(result, ofield.getValue(obj));
	}
	return result;
    }

    /**
     * Returns a full, recursive copy of the object.
     */
    public Object deepCopy(Object obj)
	   throws InstantiationException,
	          IllegalAccessException,
		  InvocationTargetException {
	return deepCopy(obj, obj.getClass());
    }

    /**
     * Returns a full, recursive copy as an instance of the specified
     * result class.  Note that this method will change the class only
     * at the top level and that leaf field values are put directly
     * into the result rather than copied.
     */
    public Object deepCopy(Object obj, Class resultClass)
	   throws InstantiationException,
	          IllegalAccessException,
		  InvocationTargetException {
	Object result = resultClass.newInstance();
	FieldWalker ow = ensureFieldWalker(obj.getClass());
	FieldWalker rw = ensureFieldWalker(resultClass);
	for (Iterator fi = ow.getFieldInfo().iterator(); fi.hasNext();) {
	    FieldWalker.FieldInfo ofield = (FieldWalker.FieldInfo)fi.next();
	    FieldWalker.FieldInfo rfield = rw.getFieldInfo(ofield.name);
	    if (rfield != null) {
		Object fvalue = ofield.getValue(obj);
		if (isLeaf(fvalue))
		    rfield.setValue(result, fvalue);
		else
		    rfield.setValue(result,
				    deepCopy(fvalue, fvalue.getClass()));
	    }
	}
	return result;
    }

    /**
     * Returns true if and only if the object does not contain fields
     * that should be recursively copied.
     */
    public boolean isLeaf(Object obj) {
	return obj == null
	    || obj instanceof SemiPrimitive
	    || obj instanceof Number
	    || obj instanceof String
	    || obj instanceof Character
	    || obj instanceof Boolean;
    }

    /**
     * Returns a FieldWalker for the class.  The walkers are cached
     * so that they needn't be computed more than once for a given
     * class.
     */
    protected FieldWalker ensureFieldWalker(Class c) {
	FieldWalker w = (FieldWalker)walkerCache.get(c);
	if (w == null) {
	    w = new FieldWalker(c);
	    walkerCache.put(c, w);
	}
	return w;
    }

}
