/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sat Jun  8 00:04:13 2002 by Jeff Dalton
 */

package ix.util.xml;

import java.util.*;

import java.io.PrintStream;

// JDOM imports
import org.jdom.output.XMLOutputter;

import ix.icore.*;
import ix.icore.domain.*;

import ix.util.*;
import ix.util.lisp.*;

/**
 * Describes the (XML) syntax of data objects.
 */
public class Syntax {

    protected XMLTranslator xmlt;

    protected XMLOutputter outputter = XML.makePrettyXMLOutputter();

    protected Map walkerCache = new HashMap();

    protected Map classToFieldsCache = new HashMap();

    protected TwoKeyHashMap fieldCaseMap =
	// Maps fieldName, containingClass -> ClassDescr
	// e.g. if the field contains the List of C, for
	// some class C, the descriptor will have C as its eltType.
	new TwoKeyHashMap();

    protected Set hiddenFields = new HashSet();

    /** Initialize tables of special cases. */
    {
	fieldCaseMap.put
	    ("nodes", Refinement.class,
	     new ClassDescr(List.class, NodeSpec.class));

	// hiddenFields.add("annotations");
    }

    public Syntax(XMLTranslator xmlt) {
	super();
	this.xmlt = xmlt;
    }

    public void describeClass(String className, PrintStream out) {
	Class c = xmlt.classForExternalName(className);
	if (c == null)
	    c = xmlt.getClassFinder().classForName(className);
	if (c == null) {
	    out.println("Can't find class named " + className);
	    return;
	}
	out.println("Syntax for " + className + " and related classes:");
	out.println("");
	List relevantClasses = relevantClasses(c);
	for (Iterator ci = relevantClasses.iterator(); ci.hasNext();) {
	    Class rc = (Class)ci.next();
	    printSyntax(rc, out);
	    out.println("");
	}
    }

    protected void printSyntax(Class c, PrintStream out) {
	ClassDescr cd = new ClassDescr(c);
	String defName = cd.getSyntaxNTName();

	out.print(defName + " ::=");

	if (cd.isStruct())
	    printStructSyntax(c, out);
	else if (cd.isEnumeration())
	    printEnumerationSyntax(c, out);

    }


    /*
     * Structure syntax
     */

    protected void printStructSyntax(Class c, PrintStream out) {
	String className = xmlt.externalNameForClass(c);
	List fields = getFields(c);
	List attrFields = attributeFields(fields);
	List eltFields = elementFields(fields);
	Debug.assert(fields.size() == attrFields.size() + eltFields.size());

	out.println("");

	out.print("   <" + className);
	if (!attrFields.isEmpty()) {
	    for (Iterator ai = attrFields.iterator(); ai.hasNext();) {
		FieldDescr fd = (FieldDescr)ai.next();
		String attrName = fd.getElementName();
		String attrValueNT = fd.type.getSyntaxNTName();
		out.println("");
		out.print("         " +
			  attrName + "=" + Strings.quote(attrValueNT));
	    }
	}
	out.println(">");

	for (Iterator ei = eltFields.iterator(); ei.hasNext();) {
	    FieldDescr fd = (FieldDescr)ei.next();
	    out.println("      " + fieldSyntax(fd));
	}

	out.println("   </" + className + ">");
    }

    protected String fieldSyntax(FieldDescr fd) {
	StringBuffer result = new StringBuffer();
	String valClassNTName = fd.type.getSyntaxNTName();
	result.append("<" + fd.getElementName() + ">");
	if (fd.type.isList()) {
	    String listName = xmlt.externalNameForClass(List.class);
	    result.append("<").append(listName).append(">")
		  .append(eltTypeSyntax(fd.type))
		  .append("</").append(listName).append(">");
	}
	else if (fd.type.isSet()) {
	    String setName = xmlt.externalNameForClass(Set.class);
	    result.append("<").append(setName).append(">")
		  .append(eltTypeSyntax(fd.type))
		  .append("</").append(setName).append(">");
	}
	else if (fd.type.isMap()) {
	    String mapName = xmlt.externalNameForClass(Map.class);
	    result.append("<").append(mapName).append(">")
		  .append("...")
		  .append("</").append(mapName).append(">");
	}
	else {
	    result.append(valClassNTName);
	}
	result.append("</" + fd.getElementName() + ">");
	return result.toString();
    }

    protected String eltTypeSyntax(ClassDescr cd) {
	String classNTName = cd.eltType == null
	    ? ""
	    : cd.eltType.getSyntaxNTName();
	return classNTName + "...";
    }

    protected List attributeFields(List fields) {
	return (List)Collect.filter(fields, isFieldAttribute);
    }

    protected List elementFields(List fields) {
	return (List)Collect.filter(fields, Fn.negate(isFieldAttribute));
    }

    protected Predicate1 isFieldAttribute = new Predicate1() {
        public boolean trueOf(Object obj) {
	    FieldDescr fd = (FieldDescr)obj;
	    return XMLTranslator.isAttributeClass(fd.type.theClass);
	}
    };

    protected void OLDprintStructSyntax(Class c, PrintStream out) {
	List fields = getFields(c);
	out.println("");
	for (Iterator fi = fields.iterator(); fi.hasNext();) {
	    FieldDescr fd = (FieldDescr)fi.next();
	    out.println("   " + fd.type + " " + fd.name);
	}
    }


    /*
     * Enumeration syntax
     */

    protected void printEnumerationSyntax(Class c, PrintStream out) {
	List values = (List)Fn.applyStatic(c, "values", new Object[]{});
	int i = 1;
	for (Iterator vi=values.iterator(); vi.hasNext(); i++) {
	    String prefix = (i == 1) ? " " : " | ";
	    out.print(prefix);
	    out.print(vi.next().toString());
	}
	out.println("");
    }

    /*
     * Finding relevant classes
     */

    /**
     * Returns a recursively composed list of classes that are related
     * to the specified class by being the types of fields, etc.
     */
    protected List relevantClasses(Class c) {
	List result = new LinkedList();
	walkRelevantClasses(c, result);
	return result;
    }

    protected void walkRelevantClasses(Class c, List result) {
	if (c == null || result.contains(c))
	    return;
	result.add(c);

	ClassDescr cd = new ClassDescr(c);
	if (cd.isStruct()) {
	    for (Iterator fi = getFields(c).iterator(); fi.hasNext();) {
		FieldDescr fd = (FieldDescr)fi.next();
		walkRelevantClasses(fd.type, result);
	    }
	}
    }

    protected void walkRelevantClasses(ClassDescr fd, List result) {
	if (fd == null)
	    return;
	else if (fd.isEnumeration() || fd.isStruct())
	    walkRelevantClasses(fd.theClass, result);
	else if (fd.isCollection())
	    walkRelevantClasses(fd.eltType, result);
	else if (fd.isMap()) {
	    walkRelevantClasses(fd.keyType, result);
	    walkRelevantClasses(fd.valType, result);
	}
    }


    /*
     * Class descriptions
     */

    protected class ClassDescr {
	Class theClass;
	ClassDescr eltType;
	ClassDescr keyType;
	ClassDescr valType;

	ClassDescr(Class theClass) {
	    this.theClass = theClass;
	}

	ClassDescr(Class theClass, Class eltClass) {
	    this.theClass = theClass;
	    this.eltType = new ClassDescr(eltClass);
	}

	ClassDescr(Class theClass, Class keyClass, Class valClass) {
	    this.theClass = theClass;
	    this.keyType = new ClassDescr(keyClass);
	    this.valType = new ClassDescr(valClass);
	}

//  	String getClassName() {
//  	    return xmlt.externalNameForClass(theClass);
//  	}

	String getElementName() {
	    return xmlt.externalNameForClass(theClass);
	}

	String getSyntaxNTName() { // nonterminal name
	    return getElementName().toUpperCase();
	}

	boolean isInterface() {
	    return theClass.isInterface();
	}

	boolean isPrimitive() {
	    return theClass.isPrimitive()
		|| theClass == String.class
		|| Number.class.isAssignableFrom(theClass)
		|| SemiPrimitive.class.isAssignableFrom(theClass);
	}

	boolean isEnumeration() {
	    return EnumeratedValue.class.isAssignableFrom(theClass);
	}

	boolean isCollection() {
	    return Collection.class.isAssignableFrom(theClass);
	}

	boolean isList() {
	    return List.class.isAssignableFrom(theClass);
	}

	boolean isSet() {
	    return Set.class.isAssignableFrom(theClass);
	}

	boolean isMap() {
	    return Map.class.isAssignableFrom(theClass);
	}

	boolean isStruct() {
	    return !(isInterface() || isPrimitive() ||
		     isCollection() || isMap());
	}

	public String toString() {
	    String className = xmlt.externalNameForClass(theClass);
	    if (isCollection()) {
		if (eltType == null)
		    return className;
		else {
		    Class eltClass = eltType.theClass;
		    String eltClassName = xmlt.externalNameForClass(eltClass);
		    return className + " [List of " + eltClassName + "]";
		}
	    }
	    else if (isMap()) {
		return className; // for now /\/
	    }
	    else {
		return className;
	    }
	}

    }


    /*
     * Field descriptions
     */

    protected class FieldDescr {
	String name;
	ClassDescr type;

	FieldDescr(String name, ClassDescr type) {
	    this.name = name;
	    this.type = type;
	}

	String getElementName() {
	    return xmlt.externalFieldName(name);
	}

    }

    /** Returns a list of FieldDescrs
     */
    protected List getFields(Class c) {
	List result = (List)classToFieldsCache.get(c);
	if (result == null) {
	    result = computeFields(c);
	    classToFieldsCache.put(c, result);
	}
	return result;
    }

    protected List computeFields(Class c) {
	List result = new LinkedList();
	for (Iterator fi = getFieldInfo(c).iterator(); fi.hasNext();) {
	    FieldWalker.FieldInfo info = (FieldWalker.FieldInfo)fi.next();
	    if (hiddenFields.contains(info.name))
		continue;
	    result.add(new FieldDescr(info.name,
				      getFieldClassDescr(info.type,
							 info.name,
							 c)));
	}
	return result;
    }

    protected ClassDescr getFieldClassDescr(Class fieldClass,
					    String fieldName,
					    Class fromClass) {
	// If fieldClass is a List or Map, we'd like to know
	// the contents types (elt, or key + val).

	// Perhaps it's a known special case.
	// /\/: Might also have a map from fieldClass.
	ClassDescr cd = (ClassDescr)fieldCaseMap.get(fieldName, fromClass);
	if (cd != null)
	    return cd;

	// For a List, we may be able to figure it out from the fieldName.
	cd = new ClassDescr(fieldClass);
	if (cd.isList() && fieldName.length() >= 2) {
	    String className = 
		Strings.capitalize
		    (fieldName.substring(0, fieldName.length() - 1));
	    // Find class from "internal" (Java) name
	    Class c = xmlt.getClassFinder().classForName(className);
	    if (c != null)
		cd.eltType = new ClassDescr(c);
	}

	// Return whatever we've managed to get
	return cd;
    }

    protected List getFieldInfo(Class c) {
	FieldWalker w = ensureFieldWalker(c);
	return w.getFieldInfo();
    }

    protected FieldWalker ensureFieldWalker(Class c) {
	FieldWalker w = (FieldWalker)walkerCache.get(c);
	if (w == null) {
	    w = new FieldWalker(c);
	    walkerCache.put(c, w);
	}
	return w;
    }


    /**
     * Test loop that repeatedly asks the user for a class name
     * and prints a description of the syntax for objects of
     * that class.
     */
    public static void main(String[] argv) {
	Syntax syntax = argv.length > 0 && argv[0].equals("java")
	    ? new Syntax(new JavaSyntaxXMLTranslator())
	    : new Syntax(new DashSyntaxXMLTranslator());
	for (;;) {
	    String in = Util.askLine("Class name:");
	    if (in.equals("bye"))
		return;
	    syntax.describeClass(in, System.out);
	    System.out.println("");
	}
    }

}

// Issues:
// * Should specify the PrintStream when constructing rather than
//   passing it from method to method?
